# deker - multidimensional arrays storage engine
# Copyright (C) 2023  OpenWeather
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from typing import TYPE_CHECKING, Generator, Optional, Union

from deker.ABC.base_managers import BaseAbstractManager, BaseManager
from deker.arrays import Array, VArray
from deker.log import SelfLoggerMixin
from deker.schemas import VArraySchema


if TYPE_CHECKING:
    from deker.ABC.base_adapters import BaseArrayAdapter, BaseVArrayAdapter
    from deker.ABC.base_schemas import BaseArraysSchema
    from deker.collection import Collection
    from deker.schemas import ArraySchema


class FilteredManager(SelfLoggerMixin, BaseAbstractManager):
    """Manager for ``Collection`` contents filtering."""

    __slots__ = (
        "__filters",
        "__collection",
        "__array_adapter",
        "__varray_adapter",
    )

    def __init__(
        self,
        collection: "Collection",
        array_adapter: "BaseArrayAdapter",
        varray_adapter: "BaseVArrayAdapter",
        schema: "BaseArraysSchema",
        filters: dict,
    ):
        super().__init__(collection, schema, array_adapter, varray_adapter)
        self.__collection = collection
        self.__schema = schema
        self.__array_adapter = array_adapter
        self.__varray_adapter = varray_adapter
        if isinstance(self.__schema, VArraySchema) and not self.__varray_adapter:
            raise AttributeError("FilterManager is missing varray adapter for varray schema")
        self.__filters = filters

    def __get_item_in_list_arrays(self, elem_index: int) -> Union[Array, VArray, None]:
        """Filter arrays and get element by index.

        :param elem_index: Which element we return
        """
        arrays = self._adapter.filter(
            self.__filters,
            self.__schema,
            self.__collection,
            self.__array_adapter,
            self.__varray_adapter,
        )
        if any(arrays):
            return arrays[elem_index]

    def first(self) -> Union["Array", "VArray", None]:
        """Return first array in the filter."""
        return self.__get_item_in_list_arrays(0)

    def last(self) -> Union["Array", "VArray", None]:
        """Return last array in the filter."""
        return self.__get_item_in_list_arrays(-1)


class DataManager(BaseManager):
    """Common data manager.

    Its behavior depends on the type of Collection (Array/Varrary)
    """

    __slots__ = (
        "__collection",
        "__array_adapter",
        "__varray_adapter",
    )

    def __init__(
        self,
        collection: "Collection",
        array_adapter: "BaseArrayAdapter",
        varray_adapter: Optional["BaseVArrayAdapter"] = None,
    ):
        schema = collection.varray_schema or collection.array_schema
        super().__init__(collection, schema, array_adapter, varray_adapter)
        self.__collection = collection
        self.__array_adapter = array_adapter
        self.__varray_adapter = varray_adapter

    def _create(  # type: ignore
        self,
        schema: "BaseArraysSchema",
        primary_attributes: Optional[dict] = None,
        custom_attributes: Optional[dict] = None,
    ) -> Union[Array, VArray]:
        """Create Array or VArray.

        :param primary_attributes: array primary attribute
        :param custom_attributes: array custom attributes
        :param schema: schema decides which array will be created
        """
        arr_params = {
            "collection": self.__collection,
            "primary_attributes": primary_attributes,
            "custom_attributes": custom_attributes,
        }

        if isinstance(schema, VArraySchema):
            arr_params.update(
                {
                    "adapter": self.__varray_adapter,  # type: ignore[dict-item]
                    "array_adapter": self.__array_adapter,  # type: ignore[dict-item]
                }
            )
        else:
            arr_params.update({"adapter": self.__array_adapter})  # type: ignore[dict-item]
        array = self._adapter.create(arr_params)
        return array

    def create(
        self, primary_attributes: Optional[dict] = None, custom_attributes: Optional[dict] = None
    ) -> Union[Array, VArray]:
        """Create array or varray.

        :param primary_attributes: Primary attributes
        :param custom_attributes:  Custom attributes
        """
        schema = self.__collection.varray_schema or self.__collection.array_schema
        return self._create(schema, primary_attributes, custom_attributes)


class VArrayManager(SelfLoggerMixin, DataManager):
    """Manager for VArrays."""

    __slots__ = (
        "__collection",
        "__array_adapter",
        "__varray_adapter",
    )

    def __init__(
        self,
        collection: "Collection",
        array_adapter: "BaseArrayAdapter",
        varray_adapter: "BaseVArrayAdapter",
    ) -> None:
        super().__init__(collection, array_adapter, varray_adapter)
        self.__collection = collection
        self.__array_adapter = array_adapter
        self.__varray_adapter = varray_adapter

    @property
    def _adapter(self) -> "BaseVArrayAdapter":
        return self.__varray_adapter  # type: ignore[return-value]

    def _get_schema(self) -> Optional["VArraySchema"]:
        return self.__collection.varray_schema

    def create(
        self,
        primary_attributes: Optional[dict] = None,
        custom_attributes: Optional[dict] = None,
    ) -> VArray:
        """Create varray in collection.

        :param primary_attributes: Primary attributes the varray
        :param custom_attributes: Custom attributes the varray
        """
        return self._create(  # type: ignore[return-value]
            self.__collection.varray_schema, primary_attributes, custom_attributes  # type: ignore[arg-type]
        )

    def __iter__(self) -> Generator[VArray, None, None]:
        """Yield VArrays from adapter."""
        self.logger.debug("iterating over VArrays")
        for meta in self._adapter:  # type: ignore[attr-defined]
            yield VArray._create_from_meta(
                self.__collection, meta, self.__array_adapter, self.__varray_adapter
            )


class ArrayManager(SelfLoggerMixin, DataManager):
    """Manager for Arrays."""

    __slots__ = (
        "__collection",
        "__array_adapter",
        "__varray_adapter",
    )

    def __init__(self, collection: "Collection", array_adapter: "BaseArrayAdapter"):
        super().__init__(collection, array_adapter, None)
        self.__collection = collection
        self.__array_adapter = array_adapter
        self.__varray_adapter = None

    @property
    def _adapter(self) -> "BaseArrayAdapter":
        return self.__array_adapter

    def _get_schema(self) -> "ArraySchema":
        """Override method which returns only array schema."""
        return self.__collection.array_schema

    def create(
        self,
        primary_attributes: Optional[dict] = None,
        custom_attributes: Optional[dict] = None,
    ) -> Array:
        """Create varray in collection.

        :param primary_attributes: Primary attributes the varray
        :param custom_attributes: Custom attributes the varray
        """
        return self._create(  # type: ignore[return-value]
            self.__collection.array_schema, primary_attributes, custom_attributes
        )

    def __iter__(self) -> Generator[Array, None, None]:
        """Yield Arrays from adapter."""
        self.logger.debug("iterating over Arrays")
        for meta in self._adapter:  # type: ignore[attr-defined]
            yield Array._create_from_meta(self.__collection, meta, self.__array_adapter)



# deker - multidimensional arrays storage engine
# Copyright (C) 2023  OpenWeather
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""Abstract interfaces for array and virtual array."""

import json

from abc import ABC, abstractmethod
from collections import OrderedDict
from copy import deepcopy
from datetime import datetime, timedelta
from typing import TYPE_CHECKING, Any, List, Optional, Tuple, Type, Union

import numpy as np

from deker_tools.slices import create_shape_from_slice
from deker_tools.time import get_utc

from deker.dimensions import Dimension, TimeDimension
from deker.errors import DekerMetaDataError, DekerValidationError
from deker.log import SelfLoggerMixin
from deker.schemas import ArraySchema, TimeDimensionSchema, VArraySchema
from deker.subset import Subset, VSubset
from deker.tools.array import check_memory, get_id
from deker.tools.schema import create_dimensions
from deker.types.private.classes import ArrayMeta, Serializer
from deker.types.private.typings import FancySlice, Numeric, Slice
from deker.validators import process_attributes


if TYPE_CHECKING:
    from deker.ABC.base_adapters import BaseArrayAdapter, BaseVArrayAdapter, IArrayAdapter
    from deker.arrays import Array, VArray
    from deker.collection import Collection


class _FancySlicer(object):
    """Converter of datetimes, floats and strings into index integers."""

    @staticmethod
    def _convert_value_for_time_dimension(
        value: FancySlice,
        dimension: TimeDimension,
    ) -> int:
        """Return value position (index) in TimeDimension.

        :param value: value to be converted into position
        :param dimension: TimeDimension instance
        """
        if isinstance(value, int):
            return value

        start: datetime = dimension.start_value
        step = dimension.step

        # check step
        if isinstance(value, timedelta):
            # if passed timedelta does not match step size
            if (value % step).total_seconds():
                raise IndexError(f"TimeDimension {dimension} wrong indexing: check 'step' value")
            return int(value // step)

        # check start and stop type
        try:
            dt = get_utc(value)  # type: ignore[arg-type]
            position = int((dt - start) // step)

            # if passed timestamp/iso-string does not match step size
            if remainder := int(((dt - start) % step).total_seconds()):
                raise IndexError(
                    f"TimeDimension {dimension.name} wrong indexing: "
                    f"{dimension.name} has no index {position}.{remainder}"
                )

            if position < 0 or position > dimension.size:
                raise IndexError(
                    f"{dimension.__class__.__name__} {dimension.name} out of range: {position}"
                )
            return position
        except (ValueError, TypeError) as e:
            raise IndexError(
                f"TimeDimension {dimension} wrong value: {value} of type {type(value)}; {e}"
            )

    @staticmethod
    def _convert_value_for_dimension(value: FancySlice, dimension: Dimension) -> int:
        """Return value position (index) in Dimension.

        :param value: value to be converted into position
        :param dimension: Dimension instance
        """
        if isinstance(value, int):
            return value

        if isinstance(value, str):
            if not dimension.labels:
                raise IndexError(
                    f"Dimension {dimension.name} wrong indexing: {dimension.name} has no labels"
                )
            dim_pos = dimension.labels.name_to_index(value)
            if dim_pos is None:
                raise IndexError(
                    f"Dimension {dimension.name} wrong indexing: {dimension.name} has no label {value}"
                )
        elif isinstance(value, float):
            if not dimension.scale:
                raise IndexError(
                    f"Dimension {dimension.name} wrong indexing: {dimension.name} has no scale"
                )

            # if passed float does not match scale step size
            if remainder := (value - dimension.scale.start_value) % dimension.scale.step:
                raise IndexError(
                    f"Dimension {dimension.name} wrong indexing: {dimension.name} has no index {remainder}"
                )

            dim_pos = int((value - dimension.scale.start_value) // dimension.scale.step)

        else:
            raise IndexError(f"Dimension {dimension.name} wrong indexing: {value} of {type(value)}")
        if dim_pos < 0 or dim_pos >= dimension.size:
            raise IndexError(f"Dimension {dimension.name} wrong indexing: {value} is out of range")
        return dim_pos

    def _convert_fancy_slice(
        self, slice_: FancySlice, dimension: Union[TimeDimension, Dimension]
    ) -> slice:
        """Convert FancySlice to slice.

        :param slice_: FancySlice to be converted into slice
        :param dimension: TimeDimension or Dimension instance
        """
        slice_parameters = []

        if all(
            st is not None for st in (slice_.start, slice_.stop)
        ) and not type(  # noqa: E721,E714
            slice_.start
        ) is type(  # noqa: E721,RUF100
            slice_.stop
        ):
            raise IndexError(
                f"'start' and 'stop' of {dimension.__class__.__name__} {dimension.name} slice "
                f"shall be of the same type"
            )

        if slice_.step is not None:
            valid_step_types = int
            if isinstance(dimension, TimeDimension):
                valid_step_types = (int, timedelta)
            if not isinstance(slice_.step, valid_step_types):
                raise IndexError(
                    f"'step' of {dimension.__class__.__name__} {dimension.name} slice "
                    f"shall be None or {valid_step_types}"
                )

        for attr in ("start", "stop", "step"):
            value = getattr(slice_, attr)

            # check if some attrs are None
            if value is None:
                if attr == "start":
                    pos = 0
                elif attr == "stop":
                    pos = dimension.size
                else:
                    pos = None
                slice_parameters.append(pos)
                continue

            if isinstance(dimension, TimeDimension):
                dim_pos = self._convert_value_for_time_dimension(value, dimension)
            else:  # Dimension
                dim_pos = self._convert_value_for_dimension(value, dimension)

            if dim_pos < 0 or dim_pos > dimension.size:
                raise IndexError(
                    f"{dimension.__class__.__name__} {dimension.name} wrong indexing: {slice_}, check '{attr}' value"
                )

            slice_parameters.append(dim_pos)

        return slice(*slice_parameters)

    @staticmethod
    def _is_time_dimension(dim: Union[TimeDimension, Dimension]) -> bool:
        return isinstance(dim, TimeDimension)

    def _fancy_slice_dimensions(
        self,
        item: FancySlice,
        dimension: Union[TimeDimension, Dimension],
    ) -> Slice:
        params = item, dimension
        if isinstance(item, slice):
            val = self._convert_fancy_slice(*params)  # type: ignore[arg-type]
        elif isinstance(item, datetime) and isinstance(dimension, TimeDimension):
            val = self._convert_value_for_time_dimension(*params)  # type: ignore[arg-type]
        elif isinstance(item, (float, str)):
            val = (
                self._convert_value_for_time_dimension(*params)  # type: ignore[arg-type]
                if self._is_time_dimension(dimension)
                else self._convert_value_for_dimension(*params)  # type: ignore[arg-type]
            )
        else:
            raise IndexError(
                f"{dimension.__class__.__name__} {dimension.name} wrong indexing: {item}"
            )

        return val

    @staticmethod
    def _check_int_or_slice_or_ellipsis(element: FancySlice) -> bool:
        types = (int, type(Ellipsis), type(None))
        if isinstance(element, types) or (
            isinstance(element, slice)
            and all(isinstance(getattr(element, attr), types) for attr in ("start", "stop", "step"))
        ):
            return True
        return False


class BaseArray(SelfLoggerMixin, Serializer, _FancySlicer, ABC):
    """Arrays abstract class providing all its inheritors with common actions and methods."""

    __slots__ = (
        "__collection",
        "__adapter",
        "__id",
        "__is_deleted",
        "custom_attributes",
        "primary_attributes",
    )

    def _validate(
        self,
        id_: Optional[str],
        primary_attributes: Optional[dict] = None,
        custom_attributes: Optional[dict] = None,
    ) -> None:
        if id_ is not None and (
            not isinstance(id_, str) or len(id_.split("-")) != 5  # noqa[PLR2004]
        ):
            raise DekerValidationError(
                f"{self.__class__.__name__} id shall be a non-empty uuid.uuid5 string or None"
            )
        for attrs in (primary_attributes, custom_attributes):
            if attrs is not None and not isinstance(attrs, dict):
                raise DekerValidationError(f"Invalid attributes type: {type(attrs)}")

    def __init__(
        self,
        collection: "Collection",
        adapter: Union["BaseArrayAdapter", "BaseVArrayAdapter"],
        array_adapter: Optional["BaseArrayAdapter"] = None,
        id_: Optional[str] = None,
        primary_attributes: Optional[dict] = None,
        custom_attributes: Optional[dict] = None,
        *args: Any,  # noqa[ARG002]
        **kwargs: Any,  # noqa[ARG002]
    ) -> None:
        """BaseArray constructor.

        Validates id, primary and custom attributes, sets attributes to empty dict if None.

        :param collection: Collection instance, to which the array is bound
        :param adapter: Array or VArray adapter instance
        :param array_adapter: Array adapter instance
        :param id_: Array uuid string
        :param primary_attributes: primary attributes keyword mapping
        :param custom_attributes: custom attributes keyword mapping
        :param args: any arguments
        :param kwargs: any keyword arguments
        """
        super().__init__()
        self.__is_deleted = False
        self._validate(id_, primary_attributes, custom_attributes)
        self.__collection: "Collection" = collection
        self.__id: str = id_ if id_ else get_id(self)
        self.__adapter = adapter
        self.__array_adapter = array_adapter

        primary_attributes, custom_attributes = process_attributes(
            self.schema, primary_attributes, custom_attributes
        )

        self.primary_attributes: OrderedDict = (
            OrderedDict({**primary_attributes}) if primary_attributes else OrderedDict()
        )
        self.custom_attributes: dict = custom_attributes if custom_attributes else {}

    def __del__(self) -> None:
        del self.__adapter
        del self.__array_adapter
        del self.__collection

    @property
    @abstractmethod
    def _adapter(self) -> "IArrayAdapter":
        """Return Array or VArray adapter."""
        pass

    @property
    def as_dict(self) -> dict:
        """Serialize self attributes into dict."""
        return {
            "id": self.id,
            "collection": self.collection,
            "dimensions": tuple(dim.as_dict for dim in self.dimensions),
            "shape": self.shape,
            "named_shape": self.named_shape,
            "primary_attributes": self.primary_attributes,
            "custom_attributes": self.custom_attributes,
        }

    @property
    @abstractmethod
    def schema(self) -> Union[ArraySchema, VArraySchema]:
        """Collection schema.

        If it is VArray - return VArraySchema; if it is Array - return ArraySchema
        """
        pass

    @property
    def id(self) -> str:
        """Get array id."""
        return self.__id

    @property
    def dtype(self) -> Type[Numeric]:
        """Get array values data type."""
        return self.schema.dtype

    @property
    def collection(self) -> str:
        """Get collection name."""
        return self.__collection.name

    @property
    def dimensions(self) -> Tuple[Union[Dimension, TimeDimension], ...]:
        """Get array dimensions."""
        return create_dimensions(
            self.schema.dimensions,  # type: ignore[arg-type]
            {**self.primary_attributes, **self.custom_attributes},
        )

    @property
    def shape(self) -> Tuple[int, ...]:
        """Get array shape."""
        return self.schema.shape

    @property
    def named_shape(self) -> Tuple[Tuple[str, int], ...]:
        """Get array shape mapping."""
        return self.schema.named_shape

    @property
    def fill_value(self) -> Union[Numeric, type(np.nan)]:  # type: ignore[valid-type]
        """Get array fill_value."""
        return self.schema.fill_value

    def update_custom_attributes(self, attributes: dict) -> None:
        """Update array custom attributes.

        This method accepts attributes that should be changed.
        Any attribute that is not in the given dict will be kept in the object attributes without any change.
        If the given dict contains any attribute that is not in the schema, DekerValidationError will be raised.

        :param attributes: attributes for updating
        """
        if not attributes:
            raise DekerValidationError("No attributes passed for update")
        for s in self.schema.dimensions:
            if (
                isinstance(s, TimeDimensionSchema)
                and isinstance(s.start_value, str)
                and s.start_value.startswith("$")
            ):
                if s.start_value[1:] in self.primary_attributes:
                    continue
                if s.start_value[1:] not in attributes:
                    for d in self.dimensions:
                        if d.name == s.name:
                            attributes[s.start_value[1:]] = d.start_value  # type: ignore[attr-defined]
            else:
                for attr in self.schema.attributes:
                    if not attr.primary and attr.name not in attributes:
                        attributes[attr.name] = self.custom_attributes[attr.name]

        process_attributes(self.schema, self.primary_attributes, attributes)
        self._adapter.update_meta_custom_attributes(self, attributes)
        self.custom_attributes = attributes
        self.logger.info(f"{self!s} custom attributes updated: {attributes}")

    def _create_meta(self) -> str:
        """Serialize array into metadata string."""
        primary_attrs, custom_attrs = deepcopy(self.primary_attributes), deepcopy(
            self.custom_attributes
        )
        for attrs in (primary_attrs, custom_attrs):
            for key, value in attrs.items():
                if isinstance(value, datetime):
                    attrs[key] = value.isoformat()
                elif isinstance(value, np.ndarray):
                    attrs[key] = value.tolist()
                elif isinstance(value, (list, tuple)):
                    elements = []
                    for element in value:
                        if isinstance(element, np.integer):
                            elements.append(int(element))
                        else:
                            elements.append(element)
                    attrs[key] = tuple(elements)
                else:
                    attrs[key] = value
        return json.dumps(
            {
                "id": self.id,
                "primary_attributes": primary_attrs,
                "custom_attributes": custom_attrs,
            }
        )

    def read_meta(self) -> Optional[ArrayMeta]:
        """Read data from array."""
        meta = self._adapter.read_meta(self)
        self.logger.info(f"{self!s} meta data read: {meta}")
        return meta

    def delete(self) -> None:
        """Delete array from storage."""
        self._adapter.delete(self)
        self.logger.info(f"{self!s} deleted")

    def _get_fancy_item(self, item: FancySlice) -> Slice:
        if isinstance(item, tuple):
            fancy_item = []
            for n, i in enumerate(item):
                dim = self.dimensions[n]
                if not self._check_int_or_slice_or_ellipsis(i):
                    i = self._fancy_slice_dimensions(i, dim)
                if isinstance(i, int) and i > dim.size:
                    raise IndexError(
                        f"{dim.__class__.__name__} dimension {dim.name} wrong indexing: {i} out of range"
                    )
                fancy_item.append(i)
            fancy_item = tuple(fancy_item)
        elif self._check_int_or_slice_or_ellipsis(item):
            fancy_item = item
        else:
            fancy_item = self._fancy_slice_dimensions(item, self.dimensions[0])

        return fancy_item  # type: ignore[return-value]

    @classmethod
    def _create_from_meta(
        cls,
        collection: "Collection",
        meta: "ArrayMeta",
        array_adapter: "BaseArrayAdapter",
        varray_adapter: Optional["BaseVArrayAdapter"] = None,
    ) -> Union["Array", "VArray"]:
        """Create Array or VArray from metadata.

        :param collection: Collection instance
        :param meta: array metadata
        :param array_adapter: Array adapter instance
        :param varray_adapter: VArray adapter instance
        """
        if varray_adapter:
            attrs_schema = collection.varray_schema.attributes
        else:
            attrs_schema = collection.array_schema.attributes
        try:
            for attr in attrs_schema:
                attributes = (
                    meta["primary_attributes"] if attr.primary else meta["custom_attributes"]
                )

                value = attributes[attr.name]

                if attr.dtype == datetime:
                    attributes[attr.name] = get_utc(value)
                if attr.dtype == tuple:
                    if (
                        attr.primary or (not attr.primary and value is not None)
                    ) and not isinstance(value, list):
                        raise DekerMetaDataError(
                            f"Collection '{collection.name}' metadata is corrupted: "
                            f"attribute '{attr.name}' has invalid type '{type(value)}'; '{attr.dtype}' expected"
                        )

                    if attr.primary or (not attr.primary and value is not None):
                        attributes[attr.name] = tuple(value)

            arr_params = {
                "collection": collection,
                "adapter": array_adapter,
                "id_": meta["id"],
                "primary_attributes": meta.get("primary_attributes"),
                "custom_attributes": meta.get("custom_attributes"),
            }
            if varray_adapter:
                arr_params.update({"adapter": varray_adapter, "array_adapter": array_adapter})
            return cls(**arr_params)  # type: ignore[arg-type,return-value]
        except (KeyError, ValueError) as e:
            raise DekerMetaDataError(f"{cls} metadata invalid/corrupted: {e}")

    def __getitem__(self, item: FancySlice) -> Union[Subset, VSubset]:
        """Get a subset from Array or VArray.

        :param item: index expression
        """
        valid_types = int, float, str, datetime, slice, tuple, type(Ellipsis)
        if (
            not isinstance(item, valid_types)
            or isinstance(item, bool)
            or item is None
            or (
                isinstance(item, tuple)
                and (
                    not all(isinstance(i, valid_types) for i in item)
                    or any(isinstance(i, bool) for i in item)
                    or any(i is None for i in item)
                )
            )
        ):
            raise IndexError(
                "Only integers, floats, strings, datetime, tuples, slices (`:`) and ellipsis (`...`) are valid indices"
            )
        if isinstance(item, tuple) and item.count(...) > 1:
            raise IndexError("An index can only have a single ellipsis (`...`)")

        item = self._get_fancy_item(item)

        shape = create_shape_from_slice(self.shape, item)
        if any(s == 0 for s in shape):
            raise IndexError(
                f"Invalid indexing: some of the axes result in zero length: subset shape {shape}, item {item}"
            )

        check_memory(shape, self.dtype, self.__adapter.ctx.config.memory_limit)
        pre_bounds = np.index_exp[item]
        bounds: Union[List, Tuple] = []
        for i in pre_bounds:
            if i is ...:
                bounds.extend(np.index_exp[:])
            else:
                bounds.append(i)
        bounds = tuple(bounds)
        if isinstance(self.schema, VArraySchema):
            subset = VSubset(
                bounds, shape, self, self.__array_adapter, self.__adapter, self.__collection  # type: ignore[arg-type]
            )
        else:
            subset = Subset(bounds, shape, self, self.__adapter)  # type: ignore[arg-type]
        self.logger.debug(f"Created subset: {subset}")
        return subset

    def __repr__(self) -> str:
        s = f"{self.__class__.__name__}(id={self.id!r}, collection={self.collection!r})"
        if self.primary_attributes:
            s = s[:-1] + f", primary_attributes={self.primary_attributes!r})"
        if self.custom_attributes:
            s = s[:-1] + f", custom_attributes={self.custom_attributes!r})"
        return s

    def __str__(self) -> str:
        return f"{self.__class__.__name__}({self.id})"



# deker - multidimensional arrays storage engine
# Copyright (C) 2023  OpenWeather
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from __future__ import annotations

import importlib
import os
import pkgutil
import warnings

from datetime import datetime
from multiprocessing import cpu_count
from pathlib import Path
from typing import TYPE_CHECKING, Any, Dict, Generator, List, Optional, Tuple, Type, Union

from deker_tools.data import convert_size_to_human
from deker_tools.path import is_path_valid
from psutil import swap_memory, virtual_memory
from tqdm import tqdm

from deker.collection import Collection
from deker.config import DekerConfig
from deker.ctx import CTX
from deker.errors import (
    DekerClientError,
    DekerCollectionNotExistsError,
    DekerMetaDataError,
    DekerValidationError,
)
from deker.integrity import IntegrityChecker
from deker.locks import META_DIVIDER
from deker.log import SelfLoggerMixin, set_logging_level
from deker.schemas import ArraySchema, VArraySchema
from deker.tools import convert_human_memory_to_bytes
from deker.types import ArrayLockMeta, CollectionLockMeta, LocksExtensions, LocksTypes, StorageSize
from deker.uri import Uri
from deker.warnings import DekerWarning


if TYPE_CHECKING:
    from concurrent.futures import ThreadPoolExecutor

    from deker.ABC.base_adapters import BaseCollectionAdapter
    from deker.ABC.base_collection import BaseCollectionOptions
    from deker.ABC.base_factory import BaseAdaptersFactory


class Client(SelfLoggerMixin):
    """Deker ``Client`` - is the first object user starts with.

    It is used for creating and getting ``Collections`` and provides connection/path to Deker collections'
    storage by `uri`. Local collection uri shall contain `file://` schema and path to the collections storage
    on local machine. Connection to the storage is provided by a client-based context, which remains
    open while the Client is open, and vice-versa: while the context is open - the Client is open too.

    ``Client`` has a context manager which opens and closes context itself::

      with Client("file://...") as client:
          ~some important job here~

    Anyway you may use ``Client`` directly::

      client = Client("file://...")
      ~some important job here~
      client.close()

    As long as ``Client`` has a context manager, its instance is reusable::

      client = Client("file://...")
      ~some important job here~
      client.close()
      with client:
          ~some important job here~
      with client:
          ~some important job here~

    Properties
    ----------
    - ``is_closed``
    - ``is_open``
    - ``meta-version``
    - ``root_path``

    API methods
    -----------
    - ``create_collection``: creates a new ``Collection`` on the storage and returns its object instance to work with.
      Requires:
        * collection **unique** name
        * an instance of ArraySchema or VArraySchema
        * chunking and compression options (optional); default is ``None``
        * type of a storage adapter (optional); default is ``HDF5StorageAdapter``

    - ``get_collection``: returns an object of ``Collection`` by a given name if such exists, otherwise - ``None``
    - ``check_integrity``: checks the integrity of embedded storage database at different levels;

      Either performs all checks and prints found errors or exit on the first error.
      The final report may be saved to file.

    - ``calculate_storage_size``: calculates size of the whole storage or of a defined ``Collection``;
    - ``close``: closes ``Client`` and its context
    - ``clear_locks``: clears all current locks within the storage or a defined ``Collection``
    - ``__enter__``: opens ``Client`` context manager
    - ``__exit__``: automatically closes ``Client`` context manager on its exit
    - ``__iter__``: iterates over all the collections within the provided uri-path, yields ``Collection`` instances.
    """

    __slots__ = (
        "__adapter",
        "__ctx",
        "__executor",
        "__storage_adapter",
        "__is_closed",
        "__config",
        "__uri",
        "__workers",
        "__factory",
    )

    __plugins: Dict[str, Type[BaseAdaptersFactory]] = {}

    def _open(self) -> None:
        """Open client.

        This method is automatically invoked on ``Client`` initiation and in context manager.
        Normally, you don't need to use it yourself.
        """
        if not self.__plugins:
            raise DekerClientError(
                "No installed adapters are found: run `pip install deker_local_adapters`"
            )

        if self.__uri.scheme not in self.__plugins:
            raise DekerClientError(
                f"Invalid uri: {self.__uri.scheme} is not supported; {self.__uri}"
            )

        if self.is_closed:
            self.__is_closed = False

            try:
                factory = self.__plugins[self.__uri.scheme]
            except AttributeError:
                raise DekerClientError(
                    f"Invalid source: installed package does not provide AdaptersFactory "
                    f"for managing uri scheme {self.__uri.scheme}"
                )

            self.__ctx = CTX(
                uri=self.__uri,
                config=self.__config,
                executor=self.__executor,
                is_closed=self.__is_closed,
                extra=self.kwargs,  # type: ignore[has-type]
            )
            self.__factory = factory(self.__ctx, self.__uri)
            self.__adapter = self.__factory.get_collection_adapter()
            self.logger.info("Client is open")

    def __get_plugins(self) -> None:
        """Get Deker adapters plugins."""
        for package in pkgutil.iter_modules():
            if package.ispkg and (
                package.name.startswith("deker_") and package.name.endswith("_adapters")
            ):
                module = importlib.import_module(package.name)
                factory = module.AdaptersFactory
                for scheme in factory.uri_schemes:
                    if scheme not in self.__plugins:
                        self.__plugins.update({scheme: factory})

    def __init__(
        self,
        uri: str = "",
        *,
        executor: Optional[ThreadPoolExecutor] = None,
        workers: Optional[int] = None,
        write_lock_timeout: int = 60,
        write_lock_check_interval: int = 1,
        loglevel: str = "ERROR",
        memory_limit: Union[int, str] = 0,
        **kwargs: Any,
    ) -> None:
        """Deker client constructor.

        :param uri: uri to Deker storage
        :param executor: external ThreadPoolExecutor instance (optional)
        :param workers: number of threads for Deker
        :param write_lock_timeout: An amount of seconds during which a parallel writing process waits for release
          of the locked file
        :param write_lock_check_interval: An amount of time (in seconds) during which a parallel writing process sleeps
          between checks for locks
        :param loglevel: Level of Deker loggers
        :param memory_limit: Limit of memory allocation per one array/subset in bytes or in human representation of
          kilobytes, megabytes or gigabytes, e.g. ``"100K"``, ``"512M"``, ``"4G"``. Human representations will be
          converted into bytes. If result is ``<= 0`` - total RAM + total swap is used

          .. note:: This parameter is used for early runtime break in case of potential memory overflow

        :param kwargs: a wildcard, reserved for any extra parameters
        """
        try:
            set_logging_level(loglevel.upper())
            self.__get_plugins()
            mem_limit = convert_human_memory_to_bytes(memory_limit)
            self.__config = DekerConfig(  # type: ignore[call-arg]
                uri=uri,
                workers=workers if workers is not None else cpu_count() + 4,
                write_lock_timeout=write_lock_timeout,
                write_lock_check_interval=write_lock_check_interval,
                loglevel=loglevel.upper(),
                memory_limit=(
                    virtual_memory().total + swap_memory().total if mem_limit <= 0 else mem_limit
                ),
            )
            self.__uri: Uri = Uri.create(self.__config.uri)
            self.__is_closed: bool = True
            self.__adapter: Optional["BaseCollectionAdapter"] = None
            self.__factory: Optional["BaseAdaptersFactory"] = None
            self.__ctx: Optional[CTX] = None
            self.__executor: Optional[ThreadPoolExecutor] = executor
            self.kwargs = kwargs
            self._open()
        except Exception as e:
            raise DekerClientError(e)

    def __del__(self) -> None:
        """Delete client."""
        self.close()

    def _validate_collection_metadata_version(self, collection_metadata: dict) -> None:
        """Validate collection metadata.

        :param collection_metadata: Dictionary with collection metadata
        """
        if (
            "metadata_version" not in collection_metadata
            or collection_metadata["metadata_version"] > self.meta_version
        ):
            raise DekerMetaDataError(
                f"Collection metadata version is not supported by this version: adapter metadata version: "
                f"{self.meta_version}, collection metadata version: {collection_metadata.get('metadata_version')}"
            )

        for n, dim in enumerate(collection_metadata["schema"]["dimensions"]):
            if dim["type"] == "generic" and "scale" not in dim:
                collection_metadata["schema"]["dimensions"][n]["scale"] = None
        collection_metadata["metadata_version"] = self.meta_version

    @property
    def meta_version(self) -> str:
        """Get actual metadata version, provided by local adapters."""
        return self.__adapter.metadata_version

    @property
    def root_path(self) -> Path:
        """Get root path to the current storage."""
        return Path(self.__adapter.uri.path) / self.__config.collections_directory

    @property
    def is_closed(self) -> bool:
        """Check client status."""
        return self.__is_closed

    @property
    def is_open(self) -> bool:
        """Check client status."""
        return not self.__is_closed

    def calculate_storage_size(self, collection_name: str = "") -> StorageSize:
        """Get the size of the storage or of a certain collection in bytes or converted to human representation.

        .. warning:: Size calculation may take a long time. Maybe you'd like to have some coffee while it's working.

        :param collection_name: Name of a ``Collection``. If not passed, the whole storage will be counted.
        """
        paths: List[Tuple[Path, str]] = []
        if collection_name:
            collection = self.get_collection(collection_name)
            if not collection:
                raise DekerCollectionNotExistsError(
                    f"Collection {collection_name} does not exist at this storage"
                )
            paths.append(
                (
                    collection.path / self.__config.array_symlinks_directory,
                    collection._storage_adapter.file_ext,
                )
            )
        else:
            paths.extend(
                [
                    (
                        collection.path / self.__config.array_symlinks_directory,
                        collection._storage_adapter.file_ext,
                    )
                    for collection in self
                ]
            )

        warning = "Size calculation may take a long time. May be, you'd like to have some coffee while it's working."
        self.logger.warning(warning)
        warnings.warn(warning, category=DekerWarning, stacklevel=0)

        size = 0

        with tqdm(paths, ncols=80, desc=str(size)) as pbar:
            for collection, ext in pbar:
                for root, _, files in os.walk(collection):  # type: ignore[type-var]
                    for file in files:  # type: str
                        if file.endswith(ext):
                            size += os.path.getsize(Path(root) / file)  # type: ignore[arg-type]
                            pbar.set_description(str(size))
                pbar.set_description(str(size))

        human = convert_size_to_human(size)
        return StorageSize(size, human)

    def close(self) -> None:
        """Close client."""
        try:
            if self.__adapter:
                self.__adapter.close()
                self.__adapter = None
            if self.__factory:
                self.__factory.close()
                self.__factory = None
            if self.__ctx:
                self.__ctx.is_closed = True
            self.__is_closed = True
        except AttributeError as e:
            self.logger.debug(f"Exception in Client.close(): {e}")
        finally:
            self.logger.info("Client is closed")

    def create_collection(
        self,
        name: str,
        schema: Union[ArraySchema, VArraySchema],  # type: ignore[arg-type]
        collection_options: Optional[BaseCollectionOptions] = None,
        storage_adapter_type: Optional[str] = None,
    ) -> Collection:
        """Create a new ``Collection`` in the database.

        :param name: Name of new ``Collection``
        :param schema: ``Array`` or ``VArray`` schema
        :param collection_options: Options for compression and chunks (if applicable)
        :param storage_adapter_type: Type of an adapter, which works with files; default is ``HDF5StorageAdapter``
        """
        if not isinstance(name, str) or not name or name.isspace():
            raise DekerValidationError(f"Collection invalid name: {name}")

        if not isinstance(schema, (ArraySchema, VArraySchema)):
            raise DekerValidationError(
                f"Invalid schema type: {type(schema)}; ArraySchema or VArraySchema expected"
            )
        # Pick correct adapter
        collection = Collection(
            name=name,
            schema=schema,
            adapter=self.__adapter,
            factory=self.__factory,
            collection_options=collection_options,
            storage_adapter=self.__adapter.get_storage_adapter(storage_adapter_type),
        )

        self.__adapter.create(collection)
        self.logger.debug(f"Collection {collection.name} created")
        return collection

    def get_collection(
        self,
        name: str,
    ) -> Optional[Collection]:
        """Get ``Collection`` from database by its name.

        :param name: Name of a ``Collection``
        """
        try:
            collection_data: dict = self.__adapter.read(name)
            self._validate_collection_metadata_version(collection_data)
            collection = self.__adapter.create_collection_from_meta(  # type: ignore[return-value]
                collection_data, self.__factory
            )
            self.logger.debug(f"Collection {name} read from meta")
            return collection  # type: ignore[return-value]
        except DekerCollectionNotExistsError:
            self.logger.info(f"Collection {name} not found")
            return None

    def collection_from_dict(self, collection_data: dict) -> Collection:
        """Create a new ``Collection`` in the database from collection metadata dictionary.

        :param collection_data: Dictionary with collection metadata
        """
        self._validate_collection_metadata_version(collection_data)
        # TODO: add tests for update logic, edge cases for ifs
        default_fields: dict = {
            "schema": {
                "fill_value": None,
                "attributes": [],
                "dimensions": {"labels": None, "scale": None},
            },
            "options": None,
        }

        for key in default_fields:
            if key not in collection_data:
                collection_data[key] = default_fields[key]

            elif isinstance(default_fields[key], dict):
                for k in default_fields[key]:
                    if k == "dimensions":
                        for n, dim in enumerate(collection_data[key][k]):
                            if dim["type"] != "time":
                                if "labels" not in dim:
                                    collection_data[key][k][n]["labels"] = None
                                if "scale" not in dim:
                                    collection_data[key][k][n]["scale"] = None

                    elif k not in collection_data[key]:
                        collection_data[key][k] = default_fields[key][k]
        collection = self.__adapter.create_collection_from_meta(  # type: ignore[return-value]
            collection_data, self.__factory
        )
        self.__adapter.create(collection)
        self.logger.debug(f"Collection {collection.name} created from dict")
        return collection  # type: ignore[return-value]

    @staticmethod
    def _iter_collection_locks(
        path: Path, collection_name: str, lock_type: Union[LocksTypes, None] = None
    ) -> list[ArrayLockMeta]:
        """Return collection ``Arrays'`` lockfiles.

        :param path: ``Collection`` path
        :param collection_name: name of a ``Collection``
          If passed - gets locks only from passed collection, else checks every collection in client
        :param lock_type: ``LocksTypes`` enum attribute, leave empty to get every lockfile
        """
        locks: list[ArrayLockMeta] = []
        for file in Path.rglob(path, "*lock"):
            if not lock_type or file.name.endswith(LocksExtensions[lock_type.name].value):
                meta = file.name.split(META_DIVIDER)
                meta_separated_numbers = 4
                if len(meta) != meta_separated_numbers:
                    # discuss other locks print format
                    continue
                if not lock_type:
                    lock_type_name = LocksExtensions(f".{file.name.split('.')[-1]}").name
                    lock_type = LocksTypes[lock_type_name]
                lock: ArrayLockMeta = {
                    "Lockfile": file.name,
                    "Collection": collection_name,
                    "Array": meta[0],
                    "ID": meta[1],
                    "PID": meta[2],
                    "TID": meta[3].split(".")[0],
                    "Type": lock_type.value,
                    "Creation": datetime.fromtimestamp(file.stat().st_ctime).isoformat(),
                }
                print(lock)
                locks.append(lock)
        return locks

    # TODO: decide on private/public method
    def _get_locks(
        self,
        collection_name: Union[str, None] = None,
        lock_type: Union[LocksTypes, None] = None,
    ) -> list[Union[CollectionLockMeta, ArrayLockMeta]]:
        """Return lockfiles of ``Collections`` and ``Arrays``.

        :param collection_name: Name of a ``Collection``
          If passed - gets locks only from passed ``Collection``, else checks every collection in the storage
        :param lock_type: ``LocksTypes`` enum attribute, leave empty to get every lockfile
        """
        locks: list[Union[CollectionLockMeta, ArrayLockMeta]] = []
        if collection_name:
            collection = self.get_collection(collection_name)
            if not collection:
                raise DekerCollectionNotExistsError(
                    f"Collection {collection_name} does not exist at this storage"
                )
            locks = self._iter_collection_locks(
                self.root_path / collection_name, collection_name, lock_type
            )
        else:
            for directory in Path(self.root_path).iterdir():
                if (
                    not lock_type
                    or lock_type == LocksTypes.collection_lock
                    and directory.is_file()
                    and directory.name.endswith(LocksExtensions.collection_lock.value)
                ):
                    lock: CollectionLockMeta = {
                        "Lockfile": directory.name,
                        "Collection": directory.name[: -len(LocksExtensions.collection_lock.value)],
                        "Type": LocksTypes.collection_lock.value,
                        "Creation": datetime.fromtimestamp(directory.stat().st_ctime).isoformat(),
                    }
                    print(lock)
                    locks.append(lock)
                locks.extend(self._iter_collection_locks(directory, directory.name, lock_type))
        return locks

    def clear_locks(self, collection_name: Union[str, None] = None) -> None:
        """Clear the readlocks of Arrays and/or VArrays.

        :param collection_name: Name of a ``Collection``.
          If passed - clears locks only in the provided collection, else clears locks in every collection in the storage
        """
        root_path = self.root_path
        files_count = 0

        if collection_name:
            collection = self.get_collection(collection_name)
            if not collection:
                raise DekerCollectionNotExistsError(
                    f"Collection {collection_name} does not exist at this storage"
                )
            root_path = root_path / collection_name

        for file in Path.rglob(root_path, "*lock"):
            if file.name.endswith(
                (LocksExtensions.array_read_lock.value, LocksExtensions.varray_lock.value)
            ):
                file.unlink()
                files_count += 1

        print(f"{files_count} files removed")

    def check_integrity(
        self,
        level: int = 1,
        stop_on_error: bool = True,
        to_file: Union[bool, Path, str] = False,
        collection: Union[str, None] = None,
    ) -> None:
        """Run storage integrity check at one of 4 levels.

        1. checks ``Collections`` integrity.  If no collection name was passed, iterates over all the
           ``Collections`` and initialises them one by one
        2. checks ``Arrays``/``VArrays`` initialization and lockfiles
        3. checks if ``Arrays``/``VArrays`` paths are valid, including symlinks
        4. checks if stored data is consistent with file-by-file one point reading

        :param collection: Name of a ``Collection``. If passed - checks only passed collection, else checks
          every collection in the storage
        :param level: Check-level
        :param stop_on_error: Flag to stop on first path or data error
        :param to_file: Dump errors in file; accepts ``True``/``False`` or a path to file. If ``True`` - dump errors
          into a default filename in the current directory; if a ``path`` to file is passed - dump errors to the file
          with a specified name and path.
        """
        errors: str = IntegrityChecker(self, self.root_path, stop_on_error, level).check(collection)
        if not errors:
            print("Integrity check found no errors")
            return
        if to_file:
            if isinstance(to_file, Path):
                filename = to_file
            elif isinstance(to_file, str):
                filename = Path(to_file)
            else:
                filename = (
                    Path(".")
                    / f"deker_integrity_report_{datetime.now().isoformat(timespec='seconds')}.txt"
                )
            is_path_valid(filename.parent.absolute())
            with filename.open("w") as f:
                f.write(errors)
            errors += f"\n\nIntegrity check logged errors in {to_file.absolute()}"
        print(errors)

    def __iter__(self) -> Generator[Collection, None, None]:
        """Iterate over all collections in the storage."""
        for meta in self.__adapter:
            collection = self.__adapter.create_collection_from_meta(
                meta, self.__factory  # type: ignore[arg-type]
            )
            yield collection

    def __enter__(self) -> Client:
        self._open()
        return self

    def __exit__(self, exc_type: Any, exc_val: Any, exc_tb: Any) -> None:
        self.close()



# deker - multidimensional arrays storage engine
# Copyright (C) 2023  OpenWeather
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from attr import asdict, dataclass, fields


@dataclass(kw_only=True)
class DekerConfig:
    """Deker application configuration.

    :param uri: Deker uri string
    :param workers: number of threads for VArray management
    :param write_lock_timeout: number of seconds for WriteLock timeout
    :param write_lock_check_interval: number of seconds for WriteLock check
    :param memory_limit: RAM size in bytes available for Deker
    :param loglevel: Deker logging level
    """

    # configurable attributes
    uri: str
    workers: int
    write_lock_timeout: int
    write_lock_check_interval: int
    memory_limit: int
    loglevel: str = "DEBUG"

    # non configurable default attributes
    collections_directory: str = "collections"
    array_data_directory: str = "array_data"
    varray_data_directory: str = "varray_data"
    array_symlinks_directory: str = "array_symlinks"
    varray_symlinks_directory: str = "varray_symlinks"

    @property
    def as_dict(self) -> dict:
        """Serialize as dict."""
        return asdict(self)

    def __attrs_post_init__(self) -> None:
        """Check if types are correct."""
        for field in fields(self.__class__):  # type: ignore[arg-type]
            default_value = getattr(self, field.name)
            if not isinstance(default_value, field.type):
                raise ValueError(f"'{field.name}' setting has wrong type")  # noqa[TRY004]



# deker - multidimensional arrays storage engine
# Copyright (C) 2023  OpenWeather
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import datetime

from typing import TYPE_CHECKING, Optional, Tuple, Union

from deker_tools.time import get_utc

from deker.errors import DekerValidationError


if TYPE_CHECKING:
    from deker.schemas import ArraySchema, AttributeSchema, VArraySchema


def process_time_dimension_attrs(attributes: dict, attr_name: str) -> datetime.datetime:
    """Validate time attribute and return its value.

    :param attributes: attributes to validate
    :param attr_name: attribute name to validate
    """
    time_attribute = attributes.get(attr_name)
    if time_attribute is None:
        raise DekerValidationError("No start value provided for time dimension")
    if time_attribute.tzinfo is None or time_attribute.tzinfo != datetime.timezone.utc:
        time_attribute = get_utc(time_attribute)
    return time_attribute


def __process_attrs(
    attrs_schema: Tuple["AttributeSchema", ...],
    attributes: dict,
    primary_attributes: dict,
    custom_attributes: dict,
) -> None:
    for attr in attrs_schema:
        if attr.primary:
            if attr.name not in attributes:
                raise DekerValidationError(f"Key attribute missing: {attr.name}")
            if not isinstance(primary_attributes[attr.name], attr.dtype):
                raise DekerValidationError(
                    f'Key attribute "{attr.name}" invalid type: {type(primary_attributes[attr.name])}; '
                    f"expected {attr.dtype}"
                )

        else:
            custom_attribute = custom_attributes.get(attr.name)
            if custom_attribute is not None and not isinstance(custom_attribute, attr.dtype):
                raise DekerValidationError(
                    f'Custom attribute "{attr.name}" invalid type {type(custom_attributes[attr.name])}; '
                    f"expected {attr.dtype}"
                )

            if custom_attribute is None:
                if attr.dtype == datetime.datetime:
                    raise DekerValidationError(f'Custom attribute "{attr.name}" cannot be None')
                custom_attributes[attr.name] = None

        if attr.dtype == datetime.datetime and attr.name in attributes:
            try:
                utc = get_utc(attributes[attr.name])
                if attr.primary:
                    primary_attributes[attr.name] = utc
                else:
                    custom_attributes[attr.name] = utc
            except (ValueError, TypeError) as e:
                raise DekerValidationError(e)


def process_attributes(
    schema: Union["ArraySchema", "VArraySchema"],
    primary_attributes: Optional[dict],
    custom_attributes: Optional[dict],
) -> Tuple[dict, dict]:
    """Validate attributes over schema.

    :param schema: ArraySchema or VArraySchema instance
    :param primary_attributes: attributes to validate
    :param custom_attributes: attributes to validate
    """
    from deker.schemas import VArraySchema

    array_type = "VArray" if isinstance(schema, VArraySchema) else "Array"

    attrs_schema = schema.attributes if schema else None

    if primary_attributes is None:
        primary_attributes = {}
    if custom_attributes is None:
        custom_attributes = {}

    if any((primary_attributes, custom_attributes)) and not attrs_schema:
        raise DekerValidationError(f"{array_type} attributes schema is missing".capitalize())

    if any(attr.primary for attr in attrs_schema) and not primary_attributes:
        raise DekerValidationError("No primary attributes provided")

    # check if attributes have unique names
    if any((primary_attributes, custom_attributes)):
        key_names = set(primary_attributes.keys())
        custom_names = set(custom_attributes.keys())
        names_intersection = custom_names.intersection(key_names)
        if names_intersection:
            raise DekerValidationError(
                f"Key and custom attributes shall not have same names; invalid names: {names_intersection}"
            )

    attributes = {**primary_attributes, **custom_attributes}

    # check extra attributes
    schema_attrs_names = {attr.name for attr in attrs_schema}
    extra_names = set(attributes.keys()).difference(schema_attrs_names)
    if extra_names:
        raise DekerValidationError(
            f"Setting additional attributes not listed in schema is not allowed. "
            f"Invalid attributes: {sorted(extra_names)}"
        )
    __process_attrs(attrs_schema, attributes, primary_attributes, custom_attributes)  # type: ignore[arg-type]
    return primary_attributes, custom_attributes



# deker - multidimensional arrays storage engine
# Copyright (C) 2023  OpenWeather
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""Abstract interfaces for managers."""

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Optional, Union

from deker.ABC.base_schemas import BaseArraysSchema
from deker.schemas import VArraySchema


if TYPE_CHECKING:
    from deker.ABC.base_adapters import BaseArrayAdapter, BaseVArrayAdapter
    from deker.ABC.base_array import BaseArray
    from deker.collection import Collection
    from deker.managers import FilteredManager


class BaseAbstractManager(ABC):
    """Data manager interface."""

    __slots__ = ("__collection", "__array_adapter", "__varray_adapter", "_schema")

    def __init__(
        self,
        collection: "Collection",
        schema: Optional[BaseArraysSchema],
        array_adapter: "BaseArrayAdapter",
        varray_adapter: Optional["BaseVArrayAdapter"] = None,
    ) -> None:
        super().__init__()
        self.__collection = collection
        self.__array_adapter = array_adapter
        self.__varray_adapter = varray_adapter
        self._schema = schema

    @property
    def _adapter(self) -> Union["BaseArrayAdapter", "BaseVArrayAdapter"]:
        """Get array adapter from factory."""
        if isinstance(self._schema, VArraySchema):
            return self.__varray_adapter  # type: ignore[return-value]
        return self.__array_adapter

    def _get_schema(self) -> BaseArraysSchema:
        """Decide which schema to use (Array or VArray)."""
        schema = (
            self.__collection.varray_schema
            if self.__collection.varray_schema
            else self.__collection.array_schema
        )
        return schema  # type: ignore[return-value]

    def filter(self, filters: dict) -> "FilteredManager":
        """Filter Arrays or VArrays by provided condition.

        :param filters: filtering parameters
        """
        from deker.managers import FilteredManager

        return FilteredManager(
            self.__collection,
            self.__array_adapter,
            self.__varray_adapter,  # type: ignore[arg-type]
            schema=self._get_schema(),
            filters=filters,
        )


class BaseManager(BaseAbstractManager):
    """Parent class for ArrayManager and VArrayManager."""

    @abstractmethod
    def create(
        self, primary_attributes: Optional[dict] = None, custom_attributes: Optional[dict] = None
    ) -> "BaseArray":
        """Create an array.

        :param primary_attributes: (V)array primary attributes
        :param custom_attributes: (V)array custom attributes
        """
        pass



# deker - multidimensional arrays storage engine
# Copyright (C) 2023  OpenWeather
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import logging
import os

from logging import Logger


_ROOT_DEKER_LOGGER_NAME = "Deker"
_level = os.getenv("DEKER_LOGLEVEL", "WARNING")
_format = "%(name)s %(levelname)-4s [%(asctime)s] %(message)s"
_logger = logging.getLogger(_ROOT_DEKER_LOGGER_NAME)
_handler = logging.StreamHandler()
_handler.setFormatter(logging.Formatter(fmt=_format))
_logger.addHandler(_handler)


class SelfLoggerMixin(object):
    """Mixin with a logger object with a possibility to log its actions."""

    __logger: Logger = None

    @property
    def logger(self) -> Logger:
        """Lazy deker logger property."""
        if not self.__logger:
            self.__logger = _logger.getChild(self.__class__.__name__)
        return self.__logger


def set_logging_level(level: str) -> None:
    """Set level of all deker loggers.

    :param level: level of logging
    """
    _logger = logging.getLogger(_ROOT_DEKER_LOGGER_NAME)
    items = list(_logger.manager.loggerDict.items())  # for Python 2 and 3
    items.sort()
    for name, logger in items:
        if _logger.name in name:
            logger.setLevel(level)



# deker - multidimensional arrays storage engine
# Copyright (C) 2023  OpenWeather
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import datetime

from typing import Any, Optional, Union

from deker.ABC.base_dimension import BaseDimension
from deker.errors import DekerValidationError
from deker.log import SelfLoggerMixin
from deker.types import IndexLabels, Labels, Scale


class Dimension(SelfLoggerMixin, BaseDimension):
    """``Dimension`` of a grid axes or any other series except time (for time series use ``TimeDimension``).

    May be used for defining the majority of parameters which may be stored in an array.
    You can use ``labels`` or ``scale`` parameter to create a mapping of some names or values to the dimension indexes.

    :param name: dimension unique name (e.g. "length")
    :param size: dimension cells quantity
    :param labels: ordered mapping of dimension values names to their indexes

      ::

        ["name1", "name2", ..., "nameN"]
        ("float1", "float2", ..., "floatN")

    :param scale: a description of dimension regular scale; default None
    """

    __slots__ = ("__name", "__size", "__step", "__labels", "__scale")

    def _validate(  # noqa: C901,RUF100
        self,
        name: str,
        size: int,
        labels: Optional[Labels] = None,
        scale: Optional[Union[dict, Scale]] = None,
        **kwargs: Any,  # noqa[ARG002]
    ) -> None:
        super()._validate(name, size)

        if all(param is None for param in (labels, scale)):
            return

        if all(param is not None for param in (labels, scale)):
            raise DekerValidationError(
                f"Invalid Dimension {name} arguments: either `labels` or `scale` or none of them should be passed, "
                f"not both"
            )

        if scale is not None:
            if not isinstance(scale, (dict, Scale)):
                raise DekerValidationError(
                    f"Scale parameter value shall be an instance of deker.types.classes.Scale, not {type(self.scale)}"
                )

            if isinstance(scale, dict):
                try:
                    scale: Scale = Scale(**scale)
                except TypeError as e:
                    raise DekerValidationError(e)

            for attr in scale.__annotations__:
                value = getattr(scale, attr)
                if attr == "name":
                    if value is not None and (
                        not isinstance(value, str) or not value or value.isspace()
                    ):
                        raise DekerValidationError(
                            f"Scale attribute '{attr}' value shall be non-empty string"
                        )

                elif not isinstance(value, float):
                    raise DekerValidationError(f"Scale attribute '{attr}' value shall be float")

            scale_end_non_inclusive = scale.start_value
            scale_end_non_inclusive += scale.step * size
            scale_size = int(round(abs((scale_end_non_inclusive - scale.start_value) / scale.step)))
            if scale_size != size:
                raise DekerValidationError(
                    f"Dimension {name} wrong scale: does not match dimension size {size}"
                )

        if labels is not None:
            common_labels_exc = (
                "Labels shall be an ordered sequence or mapping of unique elements to their "
                "indexes in the dimension axis"
            )

            if not isinstance(labels, (tuple, list, dict)):
                raise DekerValidationError(common_labels_exc)

            if isinstance(labels, (list, tuple, dict)) and not labels:
                raise DekerValidationError("Label cannot be empty, use None instead")

            if len(labels) != size:
                raise DekerValidationError("Labels quantity do not match dimension size")

            if isinstance(labels, dict):
                for key, val in labels.items():
                    if not isinstance(key, str) or not isinstance(val, int):
                        raise DekerValidationError(common_labels_exc)

            if isinstance(labels, (list, tuple)):
                if all(isinstance(el, (list, tuple)) for el in labels):
                    keys = {el[0] for el in labels}
                    vals = {el[1] for el in labels}
                    if len(keys) < len(labels) or len(vals) < len(labels):
                        raise DekerValidationError(common_labels_exc)
                    if not all(isinstance(key, str) for key in keys) or not all(
                        isinstance(val, int) for val in vals
                    ):
                        raise DekerValidationError(common_labels_exc)
                elif len(set(labels)) < len(labels) or all(
                    (
                        all(not isinstance(el, str) for el in labels),
                        all(not isinstance(el, int) for el in labels),
                        all(not isinstance(el, float) for el in labels),
                    )
                ):
                    raise DekerValidationError(common_labels_exc)

    def __init__(
        self,
        name: str,
        size: int,
        labels: Optional[Labels] = None,
        scale: Optional[dict] = None,
        **kwargs: Any,
    ) -> None:
        """Dimension constructor.

        Validates parameters, sets them to default and converts labels and scale to a corresponding type.

        :param name: dimension name (e.g. "time")
        :param size: dimension cells quantity
        :param labels: ordered mapping of dimension values names to their indexes; default None
            ["name1", "name2", ..., "nameN"] |
            ("name1", "name2", ..., "nameN") |
            [("name1", 0), ("name2", 1), ..., ("nameN", N)] |
            (("name1", 0), ("name2", 1), ..., ("nameN", N)) |
            {"name1": 0, "name2": 1, ..., "nameN": N}
        :param scale: a description of dimension regular scale; default None
        :param kwargs: any keyword arguments
        """
        super().__init__(name, size, **kwargs)
        self._validate(name, size, labels, scale)
        self.__name: str = name
        self.__size: int = size
        self.__step: int = 1  # step attribute kept for further usage unification
        self.__labels: Optional[IndexLabels] = labels if not labels else IndexLabels(labels)
        self.__scale: Optional[Scale] = scale if not scale else Scale(**scale)
        self.logger.debug(f"Dimension {name} instantiated")

    @property
    def step(self) -> int:
        """Get ``Dimension`` step."""
        return self.__step

    @property
    def name(self) -> str:
        """Get ``Dimension`` name."""
        return self.__name

    @property
    def size(self) -> int:
        """Get ``Dimension`` size."""
        return self.__size

    @property
    def labels(self) -> Optional[IndexLabels]:
        """Get ``Dimension`` labels."""
        return self.__labels

    @property
    def scale(self) -> Optional[Scale]:
        """Get ``Dimension`` regular scale."""
        return self.__scale

    @property
    def as_dict(self) -> dict:
        """Serialize ``Dimension`` into dict."""
        dic = {"name": self.name, "size": self.size, "step": self.step}
        if self.labels:
            dic["labels"] = self.labels
        if self.scale:
            dic["scale"] = self.scale._asdict() if self.scale else self.scale
        return dic

    def __len__(self) -> int:
        """Get ``Dimension`` size."""
        return self.size

    def __repr__(self) -> str:
        s = (
            f"{self.__class__.__name__}("
            f"name={self.name!r}, "
            f"size={self.size!r}, "
            f"step={self.step!r}"
            f")"
        )
        if self.labels:
            s = s[:-1] + f", labels={self.labels})"
        if self.scale:
            s = s[:-1] + f", scale={self.scale})"
        return s

    def __str__(self) -> str:
        return str(self.size)


class TimeDimension(SelfLoggerMixin, BaseDimension):
    """Special dimension for time series.

    :param name: dimension name (e.g. "time")
    :param size: dimension cells quantity

    :param start_value: time mark of the first cell (index 0) in dimension
    :param step: a ``datetime.timedelta`` step of ``TimeDimension`` series within the grid

    .. note::

      ``start_value`` shall be:
        - either ``datetime.datetime`` object with explicit ``timezone``
        - or ``datetime.isoformat()`` with explicit ``timezone``
        - or a reference to an array ``attribute`` name which shall start with ``$`` *(attribute name shall be defined
          without prefix "$")*::

            custom_attributes(
                {"my_start_value": datetime.datetime(2023, 1, 1, 0, tzinfo=timezone.utc),
                }

            TimeDimension(start_value: "$my_start_value", ...)
    """

    __slots__ = ("__name", "__size", "__start_value", "__step_label", "__step")

    def __init__(
        self,
        name: str,
        size: int,
        start_value: datetime.datetime,
        step: datetime.timedelta,
    ) -> None:
        super().__init__(name, size, **{"start_value": start_value, "step": step})  # type: ignore[arg-type]
        self.__name: str = name
        self.__size: int = size
        self.__step: datetime.timedelta = step
        self.__start_value: datetime.datetime = start_value
        self.logger.debug(f"dimension {name} instantiated")

    def _validate(
        self,
        name: str,
        size: int,
        start_value: Union[datetime.datetime, str],
        step: datetime.timedelta,
        **kwargs: Any,  # noqa[ARG002]
    ) -> None:
        super()._validate(name, size)
        if not isinstance(step, datetime.timedelta) or not step:
            raise DekerValidationError(
                'TimeDimension "step" shall be a none-zero datetime.timedelta'
            )

        if isinstance(start_value, str):
            if not start_value.startswith("$"):
                incorrect_time_dimension_error = (
                    'TimeDimension "start_value" of <str> type shall be '
                    "an iso-format datetime string with explicit tz info"
                    "or a reference to an attribute name starting with `$`"
                )
                try:
                    dt = datetime.datetime.fromisoformat(start_value)  # type: ignore[arg-type]
                    if dt.tzinfo is None:
                        raise DekerValidationError(incorrect_time_dimension_error)
                except ValueError:
                    raise DekerValidationError(incorrect_time_dimension_error)

        elif isinstance(start_value, datetime.datetime):
            if start_value.tzinfo is None:
                raise DekerValidationError(
                    'TimeDimension "start_value" of <datetime> type shall be '
                    "a datetime.datetime instance with explicit tzinfo"
                )

        else:
            raise DekerValidationError(
                'TimeDimension "start_value" shall be a datetime.datetime instance with explicit tzinfo'
                "or an iso-format datetime string with explicit tz info "
                "or a reference to an attribute name starting with `$`"
            )

    @property
    def name(self) -> str:
        """Get ``TimeDimension`` name."""
        return self.__name

    @property
    def size(self) -> int:
        """Get ``TimeDimension`` size."""
        return self.__size

    @property
    def start_value(self) -> Union[datetime.datetime, str]:
        """Get ``TimeDimension`` start value."""
        return self.__start_value

    @property
    def step(self) -> datetime.timedelta:
        """Get ``TimeDimension`` step."""
        return self.__step

    @property
    def as_dict(self) -> dict:
        """Serialize ``TimeDimension`` into dict."""
        return {
            "name": self.name,
            "size": self.size,
            "step": self.step,
            "start_value": (
                self.start_value.isoformat()
                if isinstance(self.start_value, datetime.datetime)
                else self.start_value
            ),
        }

    def __len__(self) -> int:
        """Get ``TimeDimension`` size."""
        return self.size

    def __repr__(self) -> str:
        s = (
            f"{self.__class__.__name__}("
            f"name={self.name!r}, "
            f"size={self.size!r}, "
            f"step={self.step!r}, "
            f"start_value={self.start_value!r})"
        )
        return s

    def __str__(self) -> str:
        return str(self.size)



# deker - multidimensional arrays storage engine
# Copyright (C) 2023  OpenWeather
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import uuid

from functools import singledispatch
from typing import Any, Dict, List, Tuple, Union

import numpy as np

from deker_tools.data import convert_size_to_human
from deker_tools.time import get_utc
from psutil import swap_memory, virtual_memory

from deker.errors import DekerMemoryError, DekerValidationError


def calculate_total_cells_in_array(seq: Union[Tuple[int, ...], List[int]]) -> int:
    """Get total quantity of cells in the array.

    :param seq: sequence of integers (normally array shape)
    """
    return int(np.prod(np.array(seq)))


def convert_human_memory_to_bytes(memory_limit: Union[int, str]) -> int:
    """Convert a human memory limit to bytes.

    :param memory_limit: memory limit provided by the user
    """
    bytes_ = 1024
    mapping: Dict[str, int] = {"k": bytes_, "m": bytes_**2, "g": bytes_**3}
    error = (
        f"invalid memory_limit value: {memory_limit}; expected `int` or `str` in format [number][unit] "
        f'where unit is one of ["k", "K", "m", "M", "g", "G"], e.g. "8G" or "512m"'
    )
    if not isinstance(memory_limit, (int, str)):
        raise DekerValidationError(error)

    if isinstance(memory_limit, int):
        return memory_limit

    limit, div = memory_limit[:-1], memory_limit.lower()[-1]
    try:
        int_limit: int = int(limit)
        bytes_result: int = int_limit * mapping[div]
        return bytes_result
    except Exception:
        raise DekerValidationError(error)


def check_memory(shape: tuple, dtype: type, mem_limit_from_settings: int) -> None:
    """Memory allocation checker decorator.

    Checks if it is possible to allocate memory for array/subset.

    :param shape: array or subset shape
    :param dtype: array or subset values dtype
    :param mem_limit_from_settings: deker ram limit in bytes
    """
    array_values = calculate_total_cells_in_array(shape)
    array_size_bytes = np.dtype(dtype).itemsize * array_values
    array_size_human = convert_size_to_human(array_size_bytes)

    current_limit = virtual_memory().available + swap_memory().free
    limit = min(mem_limit_from_settings, current_limit)
    limit_human = convert_size_to_human(limit)

    if array_size_bytes > limit:
        raise DekerMemoryError(
            f"Can not allocate {array_size_human} for array/subset with shape {shape} and dtype {dtype}. "
            f"Current Deker limit per array/subset is {limit_human}. Value in config: {mem_limit_from_settings}"
            f"Reduce shape or dtype of your array/subset or increase Deker RAM limit."
        )


def get_id(array: Any) -> str:
    """Generate unique id by object type and datetime.

    :param array: any object
    """
    from deker.arrays import Array, VArray

    @singledispatch
    def generate_id(arr: Any) -> str:
        """Generate unique id by object type and datetime.

        :param arr: any object
        """
        raise TypeError(f"Invalid object type: {type(arr)}")

    @generate_id.register(Array)
    def array_id(arr: Array) -> str:  # noqa[ARG001]
        """Generate id for Array.

        :param arr: Array type
        """
        return str(uuid.uuid5(uuid.NAMESPACE_X500, "array" + get_utc().isoformat()))

    @generate_id.register(VArray)
    def varray_id(arr: VArray) -> str:  # noqa[ARG001]
        """Generate id for VArray.

        :param arr: VArray type
        """
        return str(uuid.uuid5(uuid.NAMESPACE_OID, "varray" + get_utc().isoformat()))

    return generate_id(array)



# deker - multidimensional arrays storage engine
# Copyright (C) 2023  OpenWeather
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from concurrent.futures import ThreadPoolExecutor
from typing import TYPE_CHECKING, Any, Optional, Type


if TYPE_CHECKING:
    from deker.ABC.base_adapters import BaseStorageAdapter
    from deker.config import DekerConfig
    from deker.uri import Uri


class CTX:
    """Deker client context."""

    def __init__(
        self,
        uri: "Uri",
        config: "DekerConfig",
        storage_adapter: Optional[Type["BaseStorageAdapter"]] = None,
        executor: Optional[ThreadPoolExecutor] = None,
        is_closed: bool = False,
        extra: Optional[Any] = None,
    ):
        self.uri = uri
        self.storage_adapter = storage_adapter
        self.executor = executor
        self.is_closed = is_closed
        self.config = config
        self.extra: dict = extra if extra else dict()



# deker - multidimensional arrays storage engine
# Copyright (C) 2023  OpenWeather
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from __future__ import annotations

import fcntl
import os
import time

from pathlib import Path
from threading import get_native_id
from time import sleep
from typing import TYPE_CHECKING, Any, Callable, Dict, List, Optional, Sequence, Union
from uuid import uuid4

from deker.ABC.base_locks import BaseLock
from deker.errors import DekerLockError, DekerMemoryError
from deker.flock import Flock
from deker.tools.path import get_main_path
from deker.types.private.enums import LocksExtensions


if TYPE_CHECKING:
    from deker_local_adapters import LocalArrayAdapter

    from deker.arrays import Array, VArray
    from deker.types.private.classes import ArrayPositionedData

META_DIVIDER = ":"


class ReadArrayLock(BaseLock):
    """Read lock for Array."""

    ALLOWED_TYPES = ["LocalArrayAdapter"]

    def get_path(self, func_args: Sequence, func_kwargs: Dict) -> Path:
        """Get path to read-lock file.

        It's only the case for arrays, varrays don't have read locks.
        :param func_args: arguments of method call
        :param func_kwargs: keyword arguments of method call
        """
        # Get instance of the array
        array: Union[Path, Union[Array, VArray]] = (
            func_kwargs.get("array") or func_args[1]
        )  # zero arg is 'self'
        if isinstance(array, Path):
            path = array
            id_ = path.name.split(".")[0]
        else:
            id_ = array.id

        # Get file directory
        dir_path = get_main_path(id_, self.instance.collection_path / self.instance.data_dir)
        filename = (
            META_DIVIDER.join([f"{id_}", f"{uuid4()}", f"{os.getpid()}", f"{get_native_id()}"])
            + LocksExtensions.array_read_lock.value
        )
        # Create read lock file path
        path = dir_path / f"{filename}"

        self.logger.debug(f"Got path for array.id {id_} lock file: {path}")
        return path

    def check_existing_lock(self, func_args: Sequence, func_kwargs: Dict) -> None:
        """Read operations are not performed in case there are any WRITE locks.

        Array file would be Flocked for writing / updating, so we just need to try flock it once more
        :param func_args: arguments for called function.
        :param func_kwargs: keyword arguments for called function.
        """
        array: Union[Path, Union[Array, VArray]] = (
            func_kwargs.get("array") or func_args[1]
        )  # zero arg is 'self'
        if isinstance(array, Path):
            path = array
            id_ = path.name.split(".")[0]
        else:
            id_ = array.id

        dir_path = get_main_path(id_, self.instance.collection_path / self.instance.data_dir)
        path = dir_path / (id_ + self.instance.file_ext)

        try:
            with open(path, "r") as f:
                fcntl.flock(f, fcntl.LOCK_SH | fcntl.LOCK_NB)
                self.logger.debug(f"Set shared flock for {path}")

        except BlockingIOError:
            raise DekerLockError(f"Array {id_} is locked for update operation, cannot be read.")

    def acquire(self, path: Union[str, Path]) -> Any:
        """Read files will not be flocked - only created.

        :param path: Path to file which should be locked
        """
        # Create lock file
        self.lock = path
        open(path, "a").close()
        self.logger.debug(f"Acquired lock for {self.lock}")

    def release(self, e: Optional[Exception] = None) -> None:  # noqa[ARG002]
        """Release lock by deleting file.

        :param e: Exception that may have been raised.
        """
        if self.lock and self.lock.exists():
            self.lock.unlink()
            self.lock = None
            self.logger.debug(f"Released lock for {self.lock}")


class WriteArrayLock(BaseLock):
    """Write lock for arrays."""

    ALLOWED_TYPES = ["LocalArrayAdapter"]

    def get_path(self, func_args: Sequence, func_kwargs: Dict) -> Path:
        """Get path to the file for locking.

        :param func_args: arguments of method call
        :param func_kwargs: keyword arguments of method call
        """
        array = func_kwargs.get("array") or func_args[1]  # zero arg is 'self'
        dir_path = get_main_path(array.id, self.instance.collection_path / self.instance.data_dir)
        path = dir_path / (array.id + self.instance.file_ext)
        self.logger.debug(f"Got path for array.id {array.id} lock file: {path}")
        return path

    def check_existing_lock(self, func_args: Sequence, func_kwargs: Dict) -> None:
        """Check read locks before write execution.

        :param func_args: arguments of method call
        :param func_kwargs: keyword arguments of method call
        """
        # Increment write lock, to prevent more read locks coming.
        self.acquire(self.get_path(func_args, func_kwargs))

        # Check current read locks
        array = func_kwargs.get("array") or func_args[1]  # zero arg is 'self'
        dir_path = get_main_path(array.id, self.instance.collection_path / self.instance.data_dir)

        # If array belongs to varray, we should check if varray is also locked
        if array._vid:
            for file in dir_path.iterdir():
                # Skip lock from current process.
                if file.name.endswith(f"{os.getpid()}{LocksExtensions.varray_lock.value}"):
                    continue
                # If we've found another varray lock, that not from current process.
                if file.name.endswith(LocksExtensions.varray_lock.value):  # type: ignore
                    raise DekerLockError(f"Array {array} is locked with {file.name}")

        # Pattern that has to find any read locks
        glob_pattern = f"{array.id}:*{LocksExtensions.array_read_lock.value}"

        # Wait till there are no more read locks
        start_time = time.monotonic()
        while (time.monotonic() - start_time) <= self.instance.ctx.config.write_lock_timeout:
            if not list(dir_path.rglob(glob_pattern)):
                return

            sleep(self.instance.ctx.config.write_lock_check_interval)

        # If we hit the timeout, release write lock and raise DekerLockError
        self.release()
        raise DekerLockError(f"Array {array} is locked with read locks")


class WriteVarrayLock(BaseLock):
    """Write lock for VArrays."""

    ALLOWED_TYPES = ["VSubset"]

    # Locks that have been acquired by varray
    locks: List[Path] = []
    skip_lock: bool = False  # shows that we must skip this lock (e.g server adapters for subset)

    def check_type(self) -> None:
        """Check if the instance type (class) is allowed for locking."""
        # Circular import otherwise
        from deker_local_adapters.varray_adapter import LocalVArrayAdapter

        super().check_type()
        adapter = self.instance._VSubset__adapter
        is_running_on_local = isinstance(adapter, LocalVArrayAdapter)
        if not is_running_on_local:
            self.skip_lock = True

    def get_path(self, func_args: Sequence, func_kwargs: Dict) -> Optional[Path]:  # noqa[ARG002]
        """Path of json Varray file.

        :param func_args: arguments of the function that has been called.
        :param func_kwargs: keyword arguments of the function that has been called.
        """
        array = self.instance._VSubset__array
        adapter = self.instance._VSubset__adapter
        path = get_main_path(
            array.id, self.instance._VSubset__collection.path / adapter.data_dir
        ) / (array.id + adapter.file_ext)
        self.logger.debug(f"Got path for array.id {array.id} lock file: {path}")
        return path

    def check_locks_for_array_and_set_flock(self, filename: Path) -> Flock:
        """Check if there is no read lock.

        :param filename: Path to file that should be flocked
        :return:
        """
        # Check read lock first
        array_id = filename.name.split(".")[0]
        glob_pattern = f"{array_id}:*{LocksExtensions.array_read_lock.value}"
        for _ in filename.parent.rglob(glob_pattern):
            raise DekerLockError(f"Array {array_id} is locked on read")

        # Check write lock and set it
        lock = Flock(filename)
        lock.acquire()

        # Add flag that this array is locked by varray
        open(f"{filename}:{os.getpid()}{LocksExtensions.varray_lock.value}", "w").close()
        return lock

    def check_arrays_locks(
        self,
        arrays_positions: List[ArrayPositionedData],
        adapter: LocalArrayAdapter,
        varray: VArray,
    ) -> List[Path]:
        """Check all Arrays that are in current VArray.

        :param arrays_positions: Arrays' positions in VArray
        :param adapter: Array Adapter instance
        :param varray: VArray
        """
        currently_locked = []

        collection = varray._VArray__collection  # type: ignore[attr-defined]
        for array_position in arrays_positions:
            filename = adapter._get_symlink_filename(
                varray.id,
                array_position.vposition,
                collection.array_schema.primary_attributes,  # type: ignore[arg-type]
            )
            if not filename:
                continue

            # Path to the main file (not symlink)
            filename = filename.resolve()
            try:
                self.check_locks_for_array_and_set_flock(filename)
                self.locks.append(filename)
            except DekerLockError:
                currently_locked.append(filename)

        return currently_locked

    def check_existing_lock(self, func_args: Sequence, func_kwargs: Dict) -> None:  # noqa[ARG002]
        """If there are any array write/read lock, we shouldn't update varray.

        :param func_args: arguments of the function that has been called.
        :param func_kwargs: keyword arguments of the function that has been called.
        """
        adapter = self.instance._VSubset__array_adapter
        varray = self.instance._VSubset__array

        arrays_positions: List[ArrayPositionedData] = self.instance._VSubset__arrays

        # Clear links to locks
        self.locks = []
        # Locks that have been acquired by third party process

        # Iterate over Arrays in VArray and try to lock them. If locking fails - wait.
        # If it fails again - release all locks.
        currently_locked = self.check_arrays_locks(arrays_positions, adapter, varray)
        if not currently_locked and len(self.locks) == len(arrays_positions):
            return
        # Wait till there are no more read locks
        start_time = time.monotonic()
        while (time.monotonic() - start_time) <= adapter.ctx.config.write_lock_timeout:
            if not self.check_arrays_locks(arrays_positions, adapter, varray):
                return

            sleep(adapter.ctx.config.write_lock_check_interval)
        # Release all locks
        self.release()
        raise DekerLockError(f"VArray {varray} is locked")

    def release(self, e: Optional[Exception] = None) -> None:  # noqa[ARG002]
        """Release all locks.

        :param e: Exception that may have been raised.
        """
        # Release array locks
        for lock in self.locks:
            Flock(lock).release()
            Path(f"{lock}:{os.getpid()}{LocksExtensions.varray_lock.value}").unlink(missing_ok=True)
        super().release()

    @staticmethod
    def _inner_method_logic(
        lock: "WriteVarrayLock", args: Sequence, kwargs: Dict, func: Callable
    ) -> Any:
        """Logic of acquiring lock and getting result.

        :param lock: The lock that will be acquired
        :param func: decorated function
        :param args: arguments of decorated function
        :param kwargs: keyword arguments of decorated function
        """
        # If we want to skip logic of lock (e.g when we use server adapters)
        if lock.skip_lock:
            return lock.get_result(func, args, kwargs)
        return super()._inner_method_logic(lock, args, kwargs, func)


class CreateArrayLock(BaseLock):
    """Lock that we set when we want to create an array."""

    ALLOWED_TYPES = ["LocalArrayAdapter", "LocalVArrayAdapter"]

    def get_path(self, func_args: Sequence, func_kwargs: Dict) -> Path:
        """Return path to the file that should be locked.

        :param func_args: arguments for called method
        :param func_kwargs: keyword arguments for called method
        :return:
        """
        array = func_kwargs.get("array") or func_args[1]  # zero arg is 'self'

        # Get file directory path
        dir_path = get_main_path(array.id, self.instance.collection_path / self.instance.data_dir)
        filename = META_DIVIDER.join(
            [
                f"{array.id}",
                f"{uuid4()}",
                f"{os.getpid()}",
                f"{get_native_id()}",
            ]
        )
        # Create read lock file
        path = dir_path / f"{filename}{LocksExtensions.array_lock.value}"
        self.logger.debug(f"got path for array.id {array.id} lock file: {path}")
        return path

    def check_existing_lock(self, func_args: Sequence, func_kwargs: Dict) -> None:
        """Check if there is currently lock for array creating.

        :param func_args: arguments for called method
        :param func_kwargs: keyword arguments for called method
        """
        from deker.arrays import Array, VArray

        # Check current read locks
        array = func_kwargs.get("array") or func_args[1]  # zero arg is 'self'
        if isinstance(array, dict):
            adapter = array["adapter"].__class__.__name__
            if adapter not in self.ALLOWED_TYPES:
                raise DekerLockError(f"Adapter {adapter} is not allowed to create locks for arrays")

            # TODO: figure out a way to avoid constructing Array object here
            array_type = Array if adapter == self.ALLOWED_TYPES[0] else VArray
            array = array_type(**array)

        dir_path = get_main_path(array.id, self.instance.collection_path / self.instance.data_dir)

        # Pattern that has to find any create locks
        glob_pattern = f"{array.id}:*{LocksExtensions.array_lock.value}"
        for _ in dir_path.rglob(glob_pattern):
            raise DekerLockError(f"Array {array} is locked for creating")

        func_kwargs["array"] = array

    def get_result(self, func: Callable, args: Any, kwargs: Any) -> Any:
        """Call func, and get its result.

        :param func: decorated function
        :param args: arguments of decorated function
        :param kwargs: keyword arguments of decorated function
        """
        if kw_array := kwargs.pop("array"):
            args = list(args)
            # First elem is self, so for functions with array, set array as next arg.
            args[1] = kw_array
            result = func(*tuple(args), **kwargs)
        else:
            result = func(*args, **kwargs)
        return result


class UpdateMetaAttributeLock(BaseLock):
    """Lock for updating meta."""

    ALLOWED_TYPES = ["LocalArrayAdapter", "LocalVArrayAdapter"]

    def get_path(self, func_args: Sequence, func_kwargs: Dict) -> Path:
        """Return path to the file that should be locked.

        :param func_args: arguments for called method
        :param func_kwargs: keyword arguments for called method
        :return:
        """
        array = func_kwargs.get("array") or func_args[1]  # zero arg is 'self'

        # Get directory where file locates
        dir_path = get_main_path(array.id, self.instance.collection_path / self.instance.data_dir)
        path = dir_path / f"{array.id}{self.instance.file_ext}"
        self.logger.debug(f"Got path for array.id {array.id} lock file: {path}")
        return path


class CollectionLock(BaseLock):
    """Lock for collection creation."""

    ALLOWED_TYPES = ["LocalCollectionAdapter"]

    def get_path(self, func_args: Sequence, func_kwargs: Dict) -> Path:  # noqa[ARG002]
        """Return path to collection lock file.

        :param func_args: arguments for called method
        :param func_kwargs: keyword arguments for called method
        """
        collection = func_args[1]
        path = self.instance.collections_resource / (collection.name + ".lock")
        self.logger.debug(f"Got path for collection {collection.name} lock file: {path}")
        return path

    def release(self, e: Optional[Exception] = None) -> None:
        """Release Collection Lock.

        :param e: Exception that might have been raised
        """
        if self.lock:
            self.lock.release()
            if isinstance(e, DekerMemoryError):
                self.lock.file.unlink()
            self.lock = None



# deker - multidimensional arrays storage engine
# Copyright (C) 2023  OpenWeather
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

# nopycln: file
# isort: skip_file

from deker_local_adapters.storage_adapters.hdf5.hdf5_options import (
    HDF5Options,
    HDF5CompressionOpts,
)

from deker.arrays import Array, VArray
from deker.client import Client
from deker.collection import Collection
from deker.dimensions import Dimension, TimeDimension
from deker.managers import FilteredManager
from deker.schemas import (
    ArraySchema,
    AttributeSchema,
    DimensionSchema,
    TimeDimensionSchema,
    VArraySchema,
)
from deker.subset import Subset, VSubset
from deker.types.public.classes import Scale

__all__ = (
    # deker.adapters.hdf5
    "HDF5CompressionOpts",
    "HDF5Options",
    # deker.arrays
    "Array",
    "VArray",
    # deker.client
    "Client",
    # deker.collection
    "Collection",
    # deker.dimensions
    "Dimension",
    "TimeDimension",
    # deker.schemas
    "ArraySchema",
    "AttributeSchema",
    "DimensionSchema",
    "TimeDimensionSchema",
    "VArraySchema",
    # deker.subset
    "Subset",
    "VSubset",
)



# deker - multidimensional arrays storage engine
# Copyright (C) 2023  OpenWeather
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import datetime

from enum import Enum
from typing import List, Optional, Tuple, Union

import numpy as np

from attr import dataclass
from deker_tools.time import get_utc

from deker.ABC.base_schemas import BaseArraysSchema, BaseAttributeSchema, BaseDimensionSchema
from deker.errors import DekerInvalidSchemaError, DekerValidationError
from deker.log import SelfLoggerMixin
from deker.types import DimensionType, DTypeEnum, Labels, Numeric, Scale


@dataclass(repr=True)
class AttributeSchema(SelfLoggerMixin, BaseAttributeSchema):
    """Schema of an attribute.

    Describes requirements for the primary or custom attribute of Array or VArray.

    :param name: attribute name
    :param dtype: attribute data type
    :param primary: boolean flag for setting attribute as a key (``True``) or custom (``False``)
    """

    def __attrs_post_init__(self) -> None:
        """Validate after init."""
        super().__attrs_post_init__()
        try:
            self.dtype = DTypeEnum(self.dtype).value
        except (ValueError, KeyError):
            raise DekerValidationError(f"Invalid dtype value {self.dtype}")

        if not isinstance(self.primary, bool):
            raise DekerValidationError(f"Invalid primary value {self.primary}; boolean expected")
        self.logger.debug(f"{self.name} instantiated")

    @property
    def as_dict(self) -> dict:
        """Serialize Attribute schema as dict."""
        d = {"name": self.name, "primary": self.primary}
        if self.dtype == str:
            d["dtype"] = "string"
        elif self.dtype == tuple:
            d["dtype"] = "tuple"
        elif self.dtype == datetime.datetime:
            d["dtype"] = "datetime"
        else:
            try:
                d["dtype"] = DTypeEnum.get_name(DTypeEnum(self.dtype))
            except (KeyError, ValueError) as e:
                raise DekerInvalidSchemaError(
                    f'Schema "{self.__class__.__name__}" is invalid/corrupted: {e}'
                )

        return d


@dataclass(repr=True)
class DimensionSchema(SelfLoggerMixin, BaseDimensionSchema):
    """Schema of a ``Dimension`` for the majority of series except time.

    For time series use ``TimeDimensionSchema``.

    :param name: dimension unique name
    :param size: dimension cells quantity
    :param labels: Represents an ordered sequence of unique cells names or a mapping of unique cells names to their
      position (index) in dimension row. Size of such sequence or mapping shall be equal to the dimension size.

      May be useful if some data from different sources is grouped in one array, e.g.::

        DimensionSchema(
            name="weather_data",
            size=3,
            labels=["temperature", "pressure", "humidity"]
        )

      what means, that `pressure` data can be always found in such dimension at index ``1`` (not ``0``, nor ``-1``).

    :param scale: optional parameter; represents a regular scale description for dimension axis.
      For example, we describe dimensions for the Earth's latitude and longitude grades::

         dims = [
                DimensionSchema(name="y", size=721),
                DimensionSchema(name="x", size=1440),
             ]

      Such description may exist, but it's not quite sufficient. We can provide information for the grid::

        dims = [
                DimensionSchema(
                    name="y",
                    size=721,
                    scale=Scale(start_value=90, step=-0.25, name="lat")
                ),
                DimensionSchema(
                    name="x",
                    size=1440,
                    scale=Scale(start_value=-180, step=0.25, name="lon")
                ),
            ]

      This extra information permits us provide fancy indexing by lat/lon coordinates in degrees::

         EarthGridArray[1, 1] == EarthGridArray[89.75, -179.75]

    .. note::
       Parameters ``scale`` and ``labels`` provide a possibility of fancy indexing.
       If you are bored with calculating index positions, you may slice by labels instead them.

    .. attention:: Either ``scale`` or ``labels`` parameter or none of them shall be passed to the constructor.
       Not both of them.
    """

    labels: Optional["Labels"] = None
    scale: Optional[Union["Scale", dict]] = None

    def __attrs_post_init__(self) -> None:
        """Validate after init."""
        super().__attrs_post_init__()
        if self.labels and self.scale:
            raise DekerValidationError(
                f"Invalid DimensionSchema {self.name} arguments: either `labels` or `scale` or none of them should "
                f"be passed, not both"
            )
        if self.labels is not None:
            message = "Labels shall be a sequence of unique strings or a mapping of unique strings to unique ints"
            if not self.labels:
                raise DekerValidationError(message)
            if isinstance(self.labels, Scale) or not isinstance(self.labels, (dict, list, tuple)):
                raise DekerValidationError(message)
        if self.scale is not None:
            if not isinstance(self.scale, (Scale, dict)):
                raise DekerValidationError(
                    f"Scale parameter value shall be an instance of deker.types.classes.Scale "
                    f"or its dict mapping, not {type(self.scale)}"
                )
            if isinstance(self.scale, dict):
                try:
                    self.scale = Scale(**self.scale)
                except AttributeError as e:
                    raise DekerValidationError(e)
            for attr in self.scale.__annotations__:
                value = getattr(self.scale, attr)
                if attr == "name":
                    if value is not None and (
                        not isinstance(value, str) or not value or value.isspace()
                    ):
                        raise DekerValidationError(
                            f"`Scale attribute '{attr}' value shall be non-empty string"
                        )
                else:
                    if not isinstance(value, float):
                        raise DekerValidationError(f"Scale attribute '{attr}' value shall be float")

        self.logger.debug(f"{self.name} instantiated")

    @property
    def as_dict(self) -> dict:
        """Serialize DimensionSchema into dictionary."""
        d = super().as_dict
        d["labels"] = self.labels
        d["scale"] = self.scale._asdict() if self.scale else self.scale
        d["type"] = DimensionType.generic.value
        return d


@dataclass(repr=True)
class TimeDimensionSchema(SelfLoggerMixin, BaseDimensionSchema):
    """Dimension schema for time series.

    Describes data distribution within some time.

    :param name: dimension name
    :param size: dimension cells quantity
    :param start_value: time of the dimension's zero-point.
    :param step: Set a common time step for all the ``VArrays`` or ``VArrays`` with ``datetime.timedelta``
                 or its dictionary mapping.

    .. note::
       For setting a **common** start date and time for all the arrays use ``datetime.datetime`` with explicit timezone
       (``tzinfo``) or a string of ``datetime.isoformat()``.

       For setting an **individual** start date and time for each array pass a name of the attribute in the attributes
       list. Such reference shall start with ``$``, e.g. ``start_value="$my_attr_name"``.

       In this case the schema of such attribute (typed ``datetime.datetime``) must be provided by AttributeSchema
       and you shall pass ``datetime.datetime`` with explicit ``timezone`` on a ``Array`` or ``VArray`` creation to
       the correspondent attribute.
    """

    start_value: Union[datetime.datetime, str]
    step: datetime.timedelta

    def __attrs_post_init__(self) -> None:
        """Validate after init."""
        super().__attrs_post_init__()
        if not isinstance(self.step, datetime.timedelta) or not self.step:
            raise DekerValidationError(
                'TimeDimension schema "step" shall be a datetime.timedelta instance'
            )

        if isinstance(self.start_value, str):
            if not self.start_value or self.start_value.isspace():
                raise DekerValidationError(
                    'TimeDimension schema "start_value" shall be a non-empty string or a datetime.datetime instance'
                )

            if not self.start_value.startswith("$"):
                try:
                    self.start_value = get_utc(self.start_value)
                except ValueError:
                    raise DekerValidationError(
                        'TimeDimension schema "start_value" shall have a datetime.datetime type '
                        "with an explicit timezone or a string reference to a key or custom attribute name"
                    )
        elif isinstance(self.start_value, datetime.datetime):
            self.start_value = get_utc(self.start_value)
        else:
            raise DekerValidationError(
                'TimeDimension schema "start_value" shall be a datetime.datetime instance '
                "or an iso-format datetime.datetime string "
                "or a reference to an attribute name starting with `$`"
            )
        self.logger.debug(f"{self.name} instantiated")

    @property
    def as_dict(self) -> dict:
        """Serialize TimeDimensionSchema into dictionary."""
        d = super().as_dict
        d["type"] = DimensionType.time.value
        d["start_value"] = (
            self.start_value.isoformat()
            if isinstance(self.start_value, datetime.datetime)
            else self.start_value
        )
        d["step"] = {
            "days": self.step.days,
            "seconds": self.step.seconds,
            "microseconds": self.step.microseconds,
        }
        return d


def __common_arrays_attributes_post_init__(self: BaseArraysSchema) -> None:  # noqa[N807]
    """Validate attributes for ArraySchema or VArraySchema.

    :param self: ArraySchema or VArraySchema instance
    """
    if not isinstance(self.attributes, (tuple, list)):
        raise DekerValidationError("Attributes shall be a list or tuple of AttributeSchema")
    if any(not isinstance(a, AttributeSchema) for a in self.attributes):
        raise DekerValidationError("Attributes shall be a list or tuple of AttributeSchema")

    if len({attr.name for attr in self.attributes}) != len(self.attributes):
        raise DekerValidationError("Attribute name shall be unique")

    time_attribute_name = None
    for dim in self.dimensions:
        if isinstance(dim, TimeDimensionSchema) and isinstance(dim.start_value, str):
            time_attribute_name = dim.start_value[1:]

        if time_attribute_name and not any(
            attr.name == time_attribute_name and attr.dtype == datetime.datetime
            for attr in self.attributes
        ):
            raise DekerValidationError(
                f"No {time_attribute_name} attribute with dtype `datetime.datetime` is provided"
            )

    if self.attributes and isinstance(self.attributes, list):
        self.attributes = tuple(self.attributes)


@dataclass(repr=True, kw_only=True)
class VArraySchema(SelfLoggerMixin, BaseArraysSchema):
    """VArray schema - a common schema for all VArrays in Collection.

    Virtual array is an "array of arrays", or an "image of pixels".

    If we consider VArray as an image - it is split by virtual grid into tiles. In this case each tile -
    is an ordinary array.

    This schema describes the structure of the collection virtual arrays and how it is split into arrays.
    ArraySchema is automatically constructed from VArraySchema.

    :param dtype: an object, representing final data type of every array, e.g. ``int`` or ``numpy.float32``
    :param dimensions: an ordered sequence of DimensionSchemas and/or TimeDimensionSchemas;
    :param attributes: an optional sequence of AttributeSchema. If there is a TimeDimensionSchema, which ``start_value``
      parameter refers to some attribute name, attributes must contain at least such attribute schema, e.g.::

        AttributeSchema(
            name="forecast_dt",
            dtype=datetime.datetime,
            primary=False  # or True
          )

    :param vgrid: an ordered sequence of positive integers; used for splitting ``VArray`` into ordinary ``Arrays``.

      Each VArray dimension "size" shall be divided by the correspondent integer without remainders, thus an
      ``Array's`` shape is created. If there is no need to split any dimension, its vgrid positional integer
      shall be ``1``.

    :param arrays_shape: an ordered sequence of positive integers; used for setting the shape of ordinary ``Arrays``
      laying under a ``VArray``.

      Each integer in the sequence represents the total quantity of cells in the correspondent dimension
      of each ordinary ``Array``. Each ``VArray`` dimension "size" shall be divided by the correspondent integer
      without remainders, thus a ``VArray's`` vgrid is created.

    :param fill_value: an optional value for filling in empty cells;
      If ``None`` - default value for each dtype will be used. Numpy ``nan`` can be used only for floating numpy dtypes.
    """

    vgrid: Optional[Union[List[int], Tuple[int, ...]]] = None
    arrays_shape: Optional[Union[List[int], Tuple[int, ...]]] = None
    fill_value: Union[Numeric, type(np.nan), None] = None  # type: ignore[valid-type]
    attributes: Union[List[AttributeSchema], Tuple[AttributeSchema, ...]] = tuple()

    def __attrs_post_init__(self) -> None:
        """Validate schema, convert `vgrid` or `arrays_shape` to tuple and calculate the other grid splitter."""
        super().__attrs_post_init__()
        __common_arrays_attributes_post_init__(self)

        # get all grid splitters, passed by user
        splitters = {
            attr: getattr(self, attr) for attr in ("vgrid", "arrays_shape") if getattr(self, attr)
        }

        # validate found splitters; should be just one parameter
        if len(splitters) < 1:
            raise DekerValidationError("Either `vgrid` or `arrays_shape` shall be passed")
        if len(splitters) > 1:
            raise DekerValidationError("Either `vgrid` or `arrays_shape` shall be passed, not both")

        # extract grid splitter and its value
        splitter, value = tuple(splitters.items())[0]

        # validate splitter value
        if value is not None:
            if (
                not isinstance(value, (list, tuple))
                or not value
                or len(value) != len(self.dimensions)
                or not all(isinstance(item, int) for item in value)
                or any(item < 1 for item in value)
            ):
                raise DekerValidationError(
                    f"{splitter.capitalize()} shall be a list or tuple of positive integers, with total elements "
                    f"quantity equal to dimensions quantity ({len(self.dimensions)})"
                )
            for n, dim in enumerate(self.dimensions):
                remainder = dim.size % value[n]  # type: ignore[valid-type]
                if remainder != 0:
                    raise DekerValidationError(
                        f"Dimensions shall be split by {splitter} into equal parts without remainder: "
                        f"{dim.name} size % {splitter} element = "
                        f"{dim.size} % {value[n]} = {remainder}"  # type: ignore[valid-type]
                    )

            # convert splitter value to tuple
            if isinstance(value, list):
                setattr(self, splitter, tuple(value))

            # calculate second splitter name and value; set its value as tuple
            if splitter == "vgrid":
                other_splitter = "arrays_shape"
            else:
                other_splitter = "vgrid"
            setattr(
                self,
                other_splitter,
                tuple(int(i) for i in np.asarray(self.shape) // np.asarray(value)),
            )
        self.logger.debug("instantiated")

    @property
    def as_dict(self) -> dict:
        """Serialize as dict."""
        d = super().as_dict
        d["vgrid"] = self.vgrid
        return d


@dataclass(repr=True, kw_only=True)
class ArraySchema(SelfLoggerMixin, BaseArraysSchema):
    """Array schema - a common schema for all the ``Arrays`` in ``Collection``.

    It describes the structure of the collection arrays.

    :param dimensions: an ordered sequence of ``DimensionSchemas`` and/or ``TimeDimensionSchemas``
    :param dtype: an object, representing final data type of every array, e.g. ``int`` or ```numpy.float32```
    :param attributes: an optional sequence of AttributeSchema. If there is a TimeDimensionSchema, which ``start_value``
      parameter refers to some attribute name, attributes must contain at least such attribute schema, e.g.::

        AttributeSchema(
            name="forecast_dt",
            dtype=datetime.datetime,
            primary=False  # or True
          )

    :param fill_value: an optional value for filling in empty cells;
      If ``None`` - default value for each dtype will be used.
      Numpy ``nan`` can be used only for floating numpy dtypes.
    """

    fill_value: Union[Numeric, type(np.nan), None] = None  # type: ignore[valid-type]
    attributes: Union[List[AttributeSchema], Tuple[AttributeSchema, ...]] = tuple()

    def __attrs_post_init__(self) -> None:
        """Validate schema."""
        super().__attrs_post_init__()
        __common_arrays_attributes_post_init__(self)
        self.logger.debug("instantiated")


# moved here from deker.types.private.enums because of circular import
class SchemaTypeEnum(Enum):
    """Mapping of schema types to strings."""

    varray = VArraySchema
    array = ArraySchema



# deker - multidimensional arrays storage engine
# Copyright (C) 2023  OpenWeather
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from typing import TYPE_CHECKING, Optional, Tuple

from deker.ABC.base_array import BaseArray
from deker.schemas import ArraySchema, VArraySchema


if TYPE_CHECKING:
    from deker.ABC.base_adapters import BaseArrayAdapter, BaseVArrayAdapter
    from deker.collection import Collection


class Array(BaseArray):
    """``Array`` is an abstract wrapper over final low-level arrays containing data.

    ``Array`` is created by collection. It is a ``lazy`` object that does not interact with its data itself,
    but it possesses all information about the array. Interaction with data is performed by a ``Subset`` object,
    which is produced by ``Array.__getitem__()`` method.

    Properties
    ----------
    - ``as_dict``: serializes main information about array into dictionary, prepared for JSON
    - ``collection``: returns the name of ``Collection`` to which the ``Array`` is bound
    - ``dimensions``: returns a tuple of ``Array's`` dimensions as objects
    - ``dtype``: returns type of the ``Array`` data
    - ``id``: returns ``Array's`` id
    - ``schema``: returns ``Array's`` low-level schema
    - ``shape``: returns ``Array's`` shape as a tuple of dimension sizes
    - ``named_shape``: returns ``Array's`` shape as a tuple of dimension names bound to their sizes

    Attributes
    ~~~~~~~~~~
    According to the schema, arrays may have or may not have primary and/or custom attributes.
    Attributes are stored in low-level array's metadata. Primary attributes are immutable and ordered,
    custom attributes are mutable and unordered.

    - ``primary_attributes``: Primary attributes are used for ``Arrays`` filtering. It is an OrderedDict.
      If collection schema contains attributes schema with some of them defined as `primary`, you obtain
      a possibility to quickly find all ``Arrays`` which can have a certain attribute or meet same
      conditions based on primary attributes values.

      E.g., if we refer to the example above and store the information about arrays' data heights into
      primary attribute, we can create a filter ``{"height": 10}`` and find the ``Array``, which data
      was calculated at height equal to 10 meters.

    - ``custom_attributes``: Custom attributes are used to keep some unique information about the array
      or about its data. E.g. you store some weather data, which is calculated at different heights.
      If it is not important for filtering you can keep the information about certain heights in arrays'
      custom attribute, named "height". In this case you won't be able to search by this attribute,
      but still you'll be able to use this information in your calculations, checking it array by array.
      Anytime you can set this attribute to None for any array as it is not a global setting.
      But you are not able to change dtype of custom attributes. In other words you cannot provide attribute's
      dtype as ``int`` and set there any ``float`` or string or whatever.

    API Methods
    ------------
    - ``read_meta``: reads the array's metadata from storage
    - ``update_custom_attributes``: updates ``Array's`` custom attributes values
    - ``delete``: deletes ``Array`` from the storage with all its data and metadata
    - ``__getitem__``: returns a ``Subset`` instance bound to the ``Array``
    - ``__repr__``: ordinary behaviour
    - ``__str__``: ordinary behaviour

    :param collection: instance of the ``Collection``, to which the ``Array`` is bound
    :param adapter: ``Array`` adapter instance
    :param id_: array unique uuid string
    :param primary_attributes: any primary attributes keyword mapping
    :param custom_attributes: any custom attributes keyword mapping
    """

    __slots__ = (
        "__id",
        "custom_attributes",
        "primary_attributes",
    )

    def __init__(
        self,
        collection: "Collection",
        adapter: "BaseArrayAdapter",
        id_: Optional[str] = None,
        primary_attributes: Optional[dict] = None,
        custom_attributes: Optional[dict] = None,
    ) -> None:
        """Array initialization.

        :param collection: instance of the collection, to which the array is bound
        :param adapter: Array adapter instance
        :param id_: array unique uuid string
        :param primary_attributes: primary attributes keyword mapping
        :param custom_attributes: custom attributes keyword mapping
        """
        self.__collection = collection
        self.__adapter = adapter
        super().__init__(
            collection,
            self.__adapter,
            id_=id_,
            primary_attributes=primary_attributes,
            custom_attributes=custom_attributes,
        )

    @property
    def _adapter(self) -> "BaseArrayAdapter":
        return self.__adapter  # type: ignore[return-value]

    @property
    def _vid(self) -> Optional[str]:
        """Get id of the ``VArray`` to which the ``Array`` is bound."""
        if self.__collection.varray_schema:
            return self.primary_attributes.get("vid")

    @property
    def _v_position(self) -> Optional[Tuple[int, ...]]:
        """Get ``Array`` position in ``VArray``."""
        if self.__collection.varray_schema:
            vpos = self.primary_attributes.get("v_position")
            return tuple(vpos) if vpos else vpos

    @property
    def schema(self) -> ArraySchema:
        """Return ``ArraySchema`` of the Array's Collection."""
        return self.__collection.array_schema


class VArray(BaseArray):
    """Virtual array or ``VArray`` is an abstract virtual array that consists of ordinary ``Arrays``.

    ``VArray`` is an "array of arrays" or an "image of pixels". If we consider ``VArray`` as an image -
    it is split by virtual grid into tiles. In this case, each tile is an ordinary ``Array``.
    Meanwhile, for user ``VArray`` acts as an ordinary ``Array``.

    ``VArray`` is created by ``Collection``. It is a ``lazy`` object that does not interact with its data itself,
    but it possesses all information about the virtual array (and even a bit more). Interaction with data
    is performed by a VSubset object, which is produced by ``VArray.__getitem__()`` method.

    Properties
    -----------
    - ``as_dict``: serializes main information about ``VArray`` into dictionary, prepared for JSON
    - ``arrays_shape``: returns common shape of all the ``Arrays`` bound to the ``VArray``
    - ``collection``: returns the name of collection to which the ``VArray`` is bound
    - ``dimensions``: returns a tuple of ``VArray's`` dimensions as objects
    - ``dtype``: returns type of the ``VArray`` data
    - ``id``: returns virtual ``VArray's`` id
    - ``schema``: returns ``VArray's`` low-level schema
    - ``shape``: returns ``VArray's`` shape as a tuple of dimension sizes
    - ``named_shape``: returns ``VArray's`` shape as a tuple of dimension names bound to their sizes
    - ``vgrid``: returns virtual grid (a tuple of integers) by which ``VArray`` is split into ``Arrays``

    Attributes
    ~~~~~~~~~~
    According to the schema, ``VArrays`` may have or may not have primary attributes and/or custom attributes.
    Attributes are stored in ``VArray's`` metadata. Primary attributes are immutable and ordered,
    custom attributes are mutable and unordered.

    - ``primary_attributes``: Primary attributes are used for ``VArrays`` filtering. It is an OrderedDict.
      If collection schema contains attributes schema with some of them as `primary`, you obtain
      a possibility to quickly find all VArrays which can have a certain attribute or meet the same
      conditions based on primary attributes values.

      E.g., if we refer to the example above and store the information about arrays' data heights into
      primary attribute, we can create a filter ``{"height": 10}`` and find the ``VArray``, which data
      was calculated at heights equal to 10 meters.

    - ``custom_attributes``: Custom attributes are used to keep some unique information about the array
      or about its data. E.g., you store some weather data, which is calculated at different heights.
      If it is not important for filtering, you can keep the information about certain heights in arrays'
      custom attribute, named "height". In this case, you won't be able to search by this attribute,
      but still you'll be able to use this information in your calculations, checking it array by array.
      Anytime you can set this attribute to None for any array as it is not a global setting.
      But you are not able to change dtype of custom attributes. In other words, you cannot provide attribute's
      dtype as `int` and set there any `float` or string or whatever.

    API Methods
    -----------
    - ``read_meta``: reads the array's metadata from storage
    - ``update_custom_attributes``: updates ``VArray's`` custom attributes values
    - ``delete``: deletes ``VArray`` from the storage with all its data and metadata
    - ``__getitem__``: returns a ``Subset`` instance bound to the ``VArray``
    - ``__repr__``: ordinary behaviour
    - ``__str__``: ordinary behaviour

    :param collection: instance of the ``Collection``, to which the ``VArray`` is bound
    :param adapter: ``VArray`` adapter instance
    :param array_adapter: ``Array`` adapter instance
    :param id_: ``VArray`` unique uuid string
    :param primary_attributes: any primary attribute keyword mapping
    :param custom_attributes: any custom attributes keyword mapping
    """

    __slots__ = (
        "__id",
        "custom_attributes",
        "primary_attributes",
        "__adapter",
    )

    def __init__(
        self,
        collection: "Collection",
        adapter: "BaseVArrayAdapter",
        array_adapter: "BaseArrayAdapter",
        id_: Optional[str] = None,
        primary_attributes: Optional[dict] = None,
        custom_attributes: Optional[dict] = None,
    ) -> None:
        """VArray initialization.

        :param collection: instance of the collection, to which the virtual array is bound
        :param adapter: VArray adapter instance
        :param array_adapter: Array adapter instance
        :param id_: virtual array unique uuid string
        :param primary_attributes: any primary attribute keyword mapping
        :param custom_attributes: any custom attributes keyword mapping
        """
        self.__collection = collection
        self.__adapter = adapter
        self.__array_adapter = array_adapter
        super().__init__(
            collection,
            id_=id_,
            adapter=adapter,
            array_adapter=array_adapter,
            primary_attributes=primary_attributes,
            custom_attributes=custom_attributes,
        )

    @property
    def _adapter(self) -> "BaseVArrayAdapter":
        return self.__adapter  # type: ignore[return-value]

    @property
    def vgrid(self) -> Tuple[int, ...]:
        """Get ``VArray`` virtual grid."""
        return self.schema.vgrid  # type: ignore[return-value]

    @property
    def arrays_shape(self) -> Tuple[int, ...]:
        """Get the shape of inner ``Arrays``."""
        return self.schema.arrays_shape  # type: ignore[return-value]

    @property
    def as_dict(self) -> dict:
        """Serialize self into dict."""
        d = super().as_dict
        d["vgrid"] = self.vgrid
        return d

    def delete(self) -> None:
        """Delete ``VArray`` from storage."""
        self.__array_adapter.delete_all_by_vid(self.id, self.__collection)
        self._adapter.delete(self)
        self.logger.debug(f"VArray {self.id} is deleted")

    @property
    def schema(self) -> Optional[VArraySchema]:
        """Return ``VArraySchema`` of the VArray's Collection."""
        return self.__collection.varray_schema



# deker - multidimensional arrays storage engine
# Copyright (C) 2023  OpenWeather
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""Abstract interfaces for locks."""

from abc import ABC, abstractmethod
from functools import wraps
from pathlib import Path
from typing import Any, Callable, Dict, List, Optional, Sequence, Union

from deker.flock import Flock
from deker.log import SelfLoggerMixin


class BaseLock(SelfLoggerMixin, ABC):
    """Base class for read/write locks."""

    ALLOWED_TYPES: List[str] = []
    lock: Optional[Union[Flock, Path]] = None
    instance: Optional[Any] = None

    @abstractmethod
    def get_path(self, func_args: Sequence, func_kwargs: Dict) -> Optional[Path]:
        """Get path to the lock file.

        For Arrays it shall be .arrlock (array read lock) or path to the file (array write lock)
        For VArrays there are no specific locks for reading, for writing - lock on .json

        :param func_args: Arguments of method call
        :param func_kwargs: Keyword arguments of method call
        """
        pass

    def check_existing_lock(self, func_args: Sequence, func_kwargs: Dict) -> None:
        """Check if there is lock for given method.

        :param func_args: Arguments for function that has been called
        :param func_kwargs: Key word arguments for function that has been called
        """
        pass

    def check_type(self) -> None:
        """Check if the instance type (class) is allowed for locking."""
        # Workaround, because isinstance returns BaseClass
        cls: str = self.instance.__class__.__name__
        if cls not in self.ALLOWED_TYPES:
            raise TypeError(f"{cls} lock is not supported")

    def acquire(self, path: Optional[Path]) -> None:
        """Make lock using flock.

        :param path: Path to file that should be flocked
        """
        if not self.lock and path:
            self.lock = Flock(path)
            self.lock.acquire()
            self.logger.debug(f"Set flock for {path}")

    def release(self, e: Optional[Exception] = None) -> None:  # noqa[ARG002]
        """Release Flock.

        :param e: exception that might have been raised
        """
        if self.lock:
            self.lock.release()
            self.lock = None
            self.logger.debug(f"Released lock for {self.lock}")

    def get_result(self, func: Callable, args: Any, kwargs: Any) -> Any:
        """Call func, and get its result.

        :param func: decorated function
        :param args: Arguments of decorated function
        :param kwargs: Keyword arguments of decorated function
        """
        return func(*args, **kwargs)

    @staticmethod
    def _inner_method_logic(lock: "BaseLock", args: Sequence, kwargs: Dict, func: Callable) -> Any:
        """Logic of acquiring lock and getting a result.

        :param lock: lock to be acquired
        :param func: decorated function
        :param args: Arguments of decorated function
        :param kwargs: Keyword arguments of decorated function
        """
        lock.check_existing_lock(args, kwargs)
        path = lock.get_path(args, kwargs)
        lock.acquire(path)
        try:
            # Logic is in the separate method, so we can override its behavior
            # E.g., CreateArray lock has specific instructions for array attribute
            result = lock.get_result(func, args, kwargs)
        except Exception as e:
            lock.release(e)
            raise
        lock.release()

        return result

    def __call__(self, func: Callable) -> Any:
        """Create/get file to lock with Flock and lock it.

        If there is an existing lock, raise DekerLockError
        :param func: Func that would be decorated
        """

        @wraps(func)
        def inner(*args: Sequence, **kwargs: Dict[str, Any]) -> Any:
            """Inner function of lock decorator.

            :param args: Arguments of decorated function
            :param kwargs: Keyword arguments of decorated function
            """
            # To keep different references for locks
            lock = self.__class__()
            # as we wrap methods, we should have access to 'self' objects
            lock.instance = kwargs.get("self") or args[0]

            # Check that we don't have lock on anything that besides methods that require lock
            lock.check_type()

            # Logic is in the separate method, so we can override its behavior
            # E.g., in WriteVArrayLock, where we should skip the whole lock logic
            return lock._inner_method_logic(lock, args, kwargs, func)

        return inner



# deker - multidimensional arrays storage engine
# Copyright (C) 2023  OpenWeather
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import builtins

from typing import TYPE_CHECKING, Iterator, List, Optional, Tuple, Union

import numpy as np

from deker_tools.slices import create_shape_from_slice, match_slice_size, slice_converter
from numpy import ndarray

from deker.ABC.base_subset import BaseSubset
from deker.errors import DekerArrayError, DekerVSubsetError
from deker.locks import WriteVarrayLock
from deker.schemas import TimeDimensionSchema
from deker.tools import not_deleted
from deker.types.private.classes import (
    ArrayOffset,
    ArrayPosition,
    ArrayPositionedData,
    ArraysCoordinatesWithOffset,
)
from deker.types.private.typings import Arrays, ArraysCoordinates, Data, Numeric, Slice


if TYPE_CHECKING:
    from deker.ABC.base_adapters import BaseArrayAdapter, BaseVArrayAdapter
    from deker.arrays import Array, VArray
    from deker.collection import Collection
    from deker.dimensions import TimeDimension


class Subset(BaseSubset):
    """``Array`` subset.

    A subset of the ``Array`` data with set ``bounds``, ``shape``, ``dtype`` and ``fill_value``.

    It is final ``lazy`` object that can ``read``, ``update`` and ``clear`` the data within the ``Array``.
    Once created, it does not contain any data and does not access the storage until user manually invokes
    one of the subset API methods. If you need to get and manage all the data from the array you should create
    a subset with ``Array[:]`` or ``Array[...]``.

    Properties
    ----------
    - ``shape``: returns shape of the subset
    - ``bounds``: returns the bounds that were applied to the ``Array``
    - ``dtype``: returns type of the queried data
    - ``fill_value``: returns the value that fills empty cells instead of None

    API methods
    -----------

    - ``describe``: returns an OrderedDict with description of all the subset's dimensions
    - ``read``: returns a numpy ndarray with all the data from the Subset bounds

      .. note:: If the Subset or Array is empty, a numpy ndarray of ``fill_values`` will be returned

      .. warning:: Mind your RAM!

    - ``read_xarray``: returns an ``xarray.DataArray`` with data returned by `read` method and its description.

      .. attention:: Scalar data cannot be converted to ``xarray.DataArray``

    - ``update``: writes new data to the storage;

      .. note::
         * Data cannot be ``None``
         * Data ``shape`` should be equal to the ``Subset.shape``
         * Data ``dtype`` should be equal to the ``Array.dtype``

    - ``clear``: removes or resets with `fill_value` all data from the storage within the subset bounds;
    """

    __slots__ = (
        "__bounds",
        "__shape",
        "__array",
        "__adapter",
    )

    def __init__(
        self,
        slice_expression: Tuple[Union[slice, int], ...],
        shape: Tuple[int, ...],
        array: "Array",
        adapter: "BaseArrayAdapter",
    ):
        """Subset initialization.

        :param slice_expression: a slice, tuple of slices or numpy IndexExpression, created by ``Array.__getitem__()``
        :param shape: subset shape, calculated by ``Array.__getitem__()``
        :param array: an instance of ``Array`` to which the subset is bound
        :param adapter: ``Array`` adapter instance
        """
        self.__bounds = slice_expression
        self.__shape = shape
        self.__array = array
        self.__adapter = adapter
        super().__init__(slice_expression, shape, array, adapter)
        self.logger.debug(f"{self.__str__} instantiated")

    @not_deleted
    def read(self) -> Union[Numeric, np.ndarray]:
        """Read data from ``Array`` slice."""
        self.logger.debug(
            f"Trying to read data from {self.__array.id} bounds={slice_converter[self.__bounds]}"
        )
        data = self.__adapter.read_data(self.__array, self.__bounds)
        self.logger.info(f"{self!s} data read")
        return data

    @not_deleted
    def update(self, data: Data) -> None:
        """Update data in ``Array`` by slice.

        * Data cannot be ``None``
        * Data ``shape`` should be equal to the ``Subset.shape``
        * Data ``dtype`` should be equal to the ``Array.dtype``

        :param data: new data which shall match subset slicing
        """
        self.logger.debug(
            f"Trying to update data for {self.__array.id} bounds={slice_converter[self.__bounds]}"
        )
        if data is None:
            raise DekerArrayError("Updating data shall not be None")
        self.__adapter.update(self.__array, self.__bounds, data)
        self.logger.info(f"{self!s} data updated")

    @not_deleted
    def clear(self) -> None:
        """Clear data in ``Array`` by slice."""
        self.logger.debug(
            f"Trying to clear data for {self.__array.id} bounds={slice_converter[self.__bounds]}"
        )
        self.__adapter.clear(self.__array, self.__bounds)
        self.logger.info(f"{self!s} data cleared")


class VSubset(BaseSubset):
    """``VArray`` subset.

    A subset of ``VArray`` data with set ``bounds``, ``shape``, ``dtype`` and ``fill_value``.

    It is final ``lazy`` object that can ``read``, ``update`` and ``clear`` the data within the ``VArray``.
    Once created, it does not contain any data and does not access the storage until user manually invokes
    one of the virtual subset API methods. If you need to get and manage all the data from array you shall
    create a virtual subset with ``VArray[:]`` or ``VArray[...]``.

    Properties
    ----------
    - ``shape``: returns shape of the virtual subset
    - ``bounds``: returns bounds that were applied to ``VArray``
    - ``dtype``: returns type of queried data
    - ``fill_value``: returns value that fills empty cells instead of None

    API methods
    -----------

    - ``describe``: returns an OrderedDict with description of all VSubset's dimensions
    - ``read``: returns a numpy ndarray with all the data from the VSubset bounds

      .. note:: If the VSubset or VArray is empty, a numpy ndarray of ``fill_values`` will be returned

      .. warning:: Mind your RAM!

    - ``read_xarray``: returns an ``xarray.DataArray`` with data returned by `read` method and its description.

      .. attention:: Scalar data cannot be converted to ``xarray.DataArray``

    - ``update``: writes new data to the storage;

      .. note::
         * Data cannot be ``None``
         * Data ``shape`` shall be equal to the ``VSubset.shape``
         * Data ``dtype`` shall be equal to the ``VArray.dtype``

    - ``clear``: removes or resets with `fill_value` all data from the storage within the virtual subset bounds;

    :param slice_expression: a slice, tuple of slices or numpy IndexExpression, created by VArray.__getitem__()
    :param shape: subset shape, calculated by VArray.__getitem__()
    :param array: an instance of VArray to which the virtual subset is bound
    :param array_adapter: ArrayAdapter instance
    :param collection: Collection instance, to which the virtual array is bound
    """

    __slots__ = (
        "__bounds",
        "__shape",
        "__array",
        "__adapter",
        "__array_adapter",
        "__collection",
        "__arrays",
    )

    def __match_slice_exp(
        self, slice_exp: Union[slice, int, None], current_index: int, array: "VArray"
    ) -> ArraysCoordinates:
        """Create a list of vgrid indexes by expression.

        :param slice_exp: Slice expression (e.g 1,  slice(None, 2, 1), EllipsisType)
        :param current_index: current index of dimension
        :param array: Varray object
        """
        # Size of dimension
        dim_size = array.dimensions[current_index].size

        # Step in vgrid
        step = dim_size // array.vgrid[current_index]

        def get_offset(
            i: int, step: int, offset_start: int, start: int, offset_end: int, end: int
        ) -> ArrayOffset:
            """Calculate an offset for array.

            If it's first array, and there are more arrays in result, return offset from start and zero.
            If it's only one array, offset from end would be equal to the one, which was provided.
            For last array start offset is always zero, and offset form end would be equal
            to the one, which was provided. For any array in between, offsets would be zeroes.

            :param i: current array start point. i % step always equals 0
            :param offset_start: offset from start
            :param offset_end: offset from end
            :param step: step of vgrid
            :param start: start of current array( with offset)
            :param end: start of current array( with offset)
            """
            if i == start - offset_start:
                if end < i + step:
                    return {"start": offset_start, "end": step + offset_end}
                return {"start": offset_start, "end": 0}

            if i == end - offset_end - step and end < i + step:
                return {"start": 0, "end": step + offset_end}
            return {"start": 0, "end": 0}

        if type(slice_exp) is builtins.int:
            # We may receive an int.
            # In this case, the number of arrays is always 1.
            # We return it as a list of one element, as the upper function does not know anything about type
            # so just return it in the same format to avoid unnecessary checking
            return [
                [
                    ArraysCoordinatesWithOffset(
                        slice_exp // step,  # type: ignore[operator]
                        ArrayOffset(
                            start=slice_exp % step,  # type: ignore[operator]
                        ),
                    )
                ]
            ]

        # If its None at slice_exp
        # Get start of slice, and end of slice
        start, stop, _ = match_slice_size(dim_size, slice_exp)  # type: ignore[arg-type]
        # Offset from start
        offset_start = start % step
        # Offset from end (e.g. for step = 5, and stop = 10, offset would be 0
        # as slice ends on the same point as stop )
        offset_end = -(step - (stop % step)) if stop % step else 0
        # We subtract offsets to get start and end points which are divided to parts
        # by step (e.g. if there is 2 arrays, step is 20 and start is 12 and stop is 35,
        # to get the correct number of Arrays, we should start from 0, go up to 40 with given
        # step)
        return [
            [
                ArraysCoordinatesWithOffset(
                    i // step,
                    ArrayOffset(**get_offset(i, step, offset_start, start, offset_end, stop)),
                )
            ]
            for i in range(start - offset_start, stop - offset_end, step)
        ]

    def __get_arrays_for_dimension(
        self, slice_exp: List[Union[slice, int]], array: "VArray", current_index: int
    ) -> ArraysCoordinates:
        """Calculate which Arrays are in the given dimension.

        For every dimension, we calculate indexes of arrays for current dimension with offset
        using self.__match_slice_exp method, which gives us following:
        [[
          (<position in dimension>, (<offset from start>, <offset from end>)),
          (<position2 in dimension>, (<offset from start>, <offset from end>)),
          ...
        ]]
        After this, for every position, we create variations for rest of dimensions recursively.
        So [[(<position in dimension>, (<offset from start>, <offset from end>)), ...]], became
        [[
            (<position in dimension>, (<offset from start>, <offset from end>)),
            (<position in dimension2>, (<offset from start>, <offset from end>))
        ]]

        :param slice_exp: Slice expression (e.g [1, 2], [1, slice(None, 2, 1)], [EllipsisType])
        :param current_index: current index of dimension
        :param array: Varray object
        :return: [
            [
              (<position in dimension1>, (<offset from start>, <offset from end>)),
              (<position in dimension2>, (<offset from start>, <offset from end>)),
              ...
            ],
            [
              (<position2 in dimension1>, (<offset from start>, <offset from end>)),
              (<position2 in dimension2>, (<offset from start>, <offset from end>)),
              ...
            ]
        ]
        """
        # Exit point for recursion. If we are at last dimension, there is no more recursion.
        arrays = []
        if len(slice_exp) == 1:
            return self.__match_slice_exp(slice_exp[0], current_index, array)

        #  arrays_tuple contains a list with arrays positions like
        #  [[
        #    (< position in dimension >, ( < offset from start >, < offset from end >)),
        #    (< position2 in dimension >, ( < offset from start >, < offset from end >)),
        #    ...
        #  ]]
        arrays_tuple = self.__match_slice_exp(slice_exp[0], current_index, array)
        for array_tuple in arrays_tuple:
            # For every position, we create list with arrays indexes for left dimensions.
            # E.g:
            #  If the number of dimensions is 3, and the current dimension index is 0
            # It will create a list with positions for left 2 dimensions
            variations = self.__get_arrays_for_dimension(slice_exp[1:], array, current_index + 1)
            # variations example:
            # [[
            #    (< position in dimension1>, ( < offset from start >, < offset from end >)),
            #    (< position in dimension2 >, ( < offset from start >, < offset from end >))
            # ], ..]
            for variation in variations:
                # Concatenate every variation of dimension to current position
                arrays.append([*array_tuple, *variation])

        return arrays

    def __fill_slice_expression(
        self, array: "VArray", slice_exp: Tuple[Union[slice, None, int], ...]
    ) -> List[Union[slice, int]]:
        """Fill slice expression.

         The number of elements in slice_exp would be to equal number of dimensions.
        :param array: VArray instance
        :param slice_exp: tuple of ints or slices
        """
        dimension_given_length = len(slice_exp)

        slice_exp_ = []
        for i in range(len(array.dimensions)):
            slice_to_add = slice(None, None)
            if i < dimension_given_length:
                if isinstance(slice_exp[i], slice):
                    slice_to_add = slice_exp[i]
                elif isinstance(slice_exp[i], int):
                    slice_to_add = slice_exp[i]
                    if slice_exp[i] < 0:  # type: ignore[operator]
                        slice_to_add = (  # type: ignore[operator]
                            array.dimensions[i].size + slice_exp[i]  # type: ignore[operator]
                        )

            slice_exp_.append(slice_to_add)
        return slice_exp_  # type: ignore[return-value]

    def __get_array_subsets(
        self, slice_exp: Tuple[Union[slice, None, int], ...], array: "VArray"
    ) -> Arrays:
        """Calculate which Arrays are in the given subset.

        :param slice_exp: Slice expression (e.g [1, 2], [1, slice(None, 2, 1)], [EllipsisType])
        :param array: Varray instance
        """
        filled_slice_expression = self.__fill_slice_expression(array, slice_exp)
        arrays = self.__get_arrays_for_dimension(filled_slice_expression, array, 0)

        # Reformat to have position and bounds side by side
        array_positions = []

        # Check if dimension index has changed
        prev_position = [-1] * len(filled_slice_expression)
        # Sizes of previous elements
        prev_sizes = [0] * len(filled_slice_expression)
        # Buffer of sizes, writes to prev sizes, only for dimension that has changed
        prev_sizes_buffer = [0] * len(filled_slice_expression)

        def calc_data_slice(
            current_dimension: int, dim: Union[slice, int], array_vposition: list
        ) -> Union[slice, int, None]:
            """Calculate slice for data part.

            :param current_dimension: Index of current dimension
            :param dim: Subset dimension (e.g slice(0, 5))
            :param array_vposition: Position of array
            """
            if isinstance(filled_slice_expression[current_dimension], int):
                prev_position[current_dimension] = array_vposition[current_dimension]
                return None

            if isinstance(dim, int):
                prev_position[current_dimension] = array_vposition[current_dimension]
                return dim

            dim_start, dim_stop, _ = match_slice_size(
                array.dimensions[current_dimension].size // array.vgrid[current_dimension], dim
            )
            positions_length = len(array_vposition)

            if prev_position[current_dimension] == array_vposition[current_dimension]:
                prev_sizes_buffer[current_dimension] = prev_sizes[current_dimension]
            if prev_position[current_dimension] != array_vposition[current_dimension]:
                prev_sizes[current_dimension] = prev_sizes_buffer[current_dimension]

            for i in range(positions_length):
                if prev_position[i] != array_vposition[i] and i < positions_length - 1:
                    prev_sizes_buffer[i + 1 : :] = [0] * (positions_length - i - 1)

            elems_in_dimension = abs(dim_stop - dim_start)
            if array_vposition[current_dimension] == 0:
                prev_sizes[current_dimension] = prev_sizes_buffer[current_dimension] = 0

            slice_start = prev_sizes[current_dimension]
            slice_stop = slice_start + elems_in_dimension
            # what was prev position in that dimension
            prev_position[current_dimension] = array_vposition[current_dimension]
            # how many elements in that dimension
            prev_sizes_buffer[current_dimension] += elems_in_dimension
            return slice(slice_start, slice_stop)

        for array_ in arrays:
            vposition: List[int] = []
            bounds: List[Union[slice, int]] = []
            for index, (position, offset) in enumerate(array_):
                vposition.append(position)
                if "end" not in offset:
                    bounds.append(offset["start"])
                    continue
                start, end = offset["start"], offset["end"]
                if end == 0:
                    end = array.dimensions[index].size // array.vgrid[index]
                    if start == 0:
                        bounds.append(slice(None, None))
                        continue
                bounds.append(slice(start, end))

            dt_slice = [calc_data_slice(i, dim, vposition) for i, dim in enumerate(bounds)]
            array_positions.append(
                ArrayPosition(
                    vposition=tuple(vposition),
                    bounds=tuple(bounds),
                    data_slice=tuple([dt for dt in dt_slice if dt is not None]),
                )
            )

        sorted_positions = sorted(array_positions, key=lambda p: p.vposition)
        return sorted_positions

    def __init__(
        self,
        slice_expression: Tuple[Union[slice, int], ...],
        shape: Tuple[int, ...],
        array: "VArray",
        array_adapter: "BaseArrayAdapter",
        varray_adapter: "BaseVArrayAdapter",
        collection: "Collection",
    ):
        """VSubset constructor.

        Calculates arrays for future reading.

        :param slice_expression: a slice, tuple of slices or numpy IndexExpression, created by VArray.__getitem__()
        :param shape: subset shape, calculated by VArray.__getitem__()
        :param array: an instance of VArray to which the virtual subset is bound
        :param array_adapter: ArrayAdapter instance
        :param varray_adapter: VArrayAdapter instances
        :param collection: Collection instance
        """
        self.__bounds = slice_expression
        self.__shape = shape
        self.__array = array
        self.__adapter: "BaseVArrayAdapter" = varray_adapter
        self.__array_adapter: "BaseArrayAdapter" = array_adapter
        self.__arrays: Arrays = self.__get_array_subsets(
            slice_expression, array  # type: ignore[arg-type]
        )
        self.__collection: "Collection" = collection
        super().__init__(slice_expression, shape, array, array_adapter, varray_adapter, collection)
        self.logger.debug(f"{self.__str__} instantiated")

    def _create_array_from_vposition(self, vpos: Tuple[int, ...]) -> Optional["Array"]:
        array = self.__collection.arrays.filter({"vid": self.__array.id, "v_position": vpos}).last()
        if not array:
            self.logger.debug(f"Array for v_position {vpos} not found")
            return None
        self.logger.debug(f"Created Array from meta for v_position {vpos}: {array.id}")
        return array  # type: ignore[return-value]

    @not_deleted
    @WriteVarrayLock()
    def clear(self) -> None:
        """Clear data in ``VArray`` by slice."""

        def _clear(array_pos: ArrayPosition) -> None:
            array = self._create_array_from_vposition(array_pos.vposition)
            if array:
                subset = array[array_pos.bounds]
                if subset.shape == array.shape:
                    array.delete()
                else:
                    subset.clear()

        self.logger.debug(f"Trying to clear data for {self.__str__()}")
        results = self.__adapter.executor.map(_clear, self.__arrays)
        list(results)
        self.logger.info(f"{self!s} data cleared")

    def __sum_results(self, arrays_data: Iterator) -> np.ndarray:
        """Arrange data from arrays into VSubset shape.

        :param arrays_data: tuple of data positions in VSubset and data
        """
        results = np.empty(shape=self.shape, dtype=self.__array.dtype)
        for position, data in arrays_data:
            if position == tuple():
                return data[position]
            results[position] = data
        return results

    @not_deleted
    def read(self) -> Union[Numeric, np.ndarray]:
        """Read data from ``VArray`` slice."""

        def _read_data(array_pos: ArrayPosition) -> Tuple[Slice, Union[Numeric, ndarray, None]]:
            array: "Array" = self._create_array_from_vposition(array_pos.vposition)
            if array:
                subset: Subset = array[array_pos.bounds]
                result = subset.read()
            else:
                result = np.empty(
                    shape=create_shape_from_slice(
                        self.__array.arrays_shape, array_pos.bounds  # type: ignore[attr-defined]
                    ),
                    dtype=self.__array.dtype,
                )
                result.fill(self.__array.fill_value)
            return array_pos.data_slice, result

        self.logger.debug(f"Trying to read data from {self!s}")
        arrays_data = self.__adapter.executor.map(_read_data, self.__arrays)
        data = self.__sum_results(arrays_data)
        self.logger.info(f"{self!s} data read")
        return data

    @not_deleted
    @WriteVarrayLock()
    def update(self, data: Data) -> None:
        """Update data in ``VArray`` by slice.

        * Data cannot be ``None``
        * Data ``shape`` should be equal to the ``VSubset.shape``
        * Data ``dtype`` should be equal to the ``VArray.dtype``

        :param data: new data which shall match subset slicing
        """

        def _update(array_data: ArrayPositionedData) -> None:
            """If there is a need in the future to calculate Array's time dimension start value.  # noqa: DAR101, D400

            ATD - Array time dimension
            VATD - VArray time dimension
            vpos - v_position

            start_value = VATD.step * ATD.size * vpos[vpos.index(ATD)] + VATD.start_value
            """
            array = self._create_array_from_vposition(array_data.vposition)
            if not array:
                custom_attributes = {}
                for n, dim_schema in enumerate(self.__collection.array_schema.dimensions):
                    if isinstance(dim_schema, TimeDimensionSchema) and isinstance(
                        dim_schema.start_value, str
                    ):
                        attr_name = dim_schema.start_value[1:]
                        dim: TimeDimension = self.__array.dimensions[n]
                        pos = array_data.vposition[n]
                        custom_attributes[attr_name] = dim.start_value + dim.step * pos  # type: ignore[operator]

                array = self.__array_adapter.create(
                    {
                        "collection": self.__collection,
                        "adapter": self.__array_adapter,
                        "primary_attributes": {
                            "vid": self.__array.id,
                            "v_position": array_data.vposition,
                        },
                        "custom_attributes": custom_attributes,
                    }
                )
            subset = array[array_data.bounds]
            subset.update(array_data.data)

        self.logger.debug(f"Trying to update data for {self!s}")
        if data is None:
            raise DekerArrayError("Updating data shall not be None")

        data = self.__array_adapter._process_data(
            self.__array.dtype, self.__array.shape, data, self.__bounds
        )

        results = self.__adapter.executor.map(
            _update,
            [
                ArrayPositionedData(vpos, array_bounds, data[data_bounds])
                for vpos, array_bounds, data_bounds in self.__arrays
            ],
        )
        try:
            list(results)
        except Exception as e:
            raise DekerVSubsetError(
                f"ATTENTION: Data in {self!s} IS NOW CORRUPTED due to the exception above"
            ).with_traceback(e.__traceback__)
        self.logger.info(f"{self!s} data updated OK")
